"""Input data for surfex binaries."""

import json
import logging
import os
import subprocess
from abc import ABC, abstractmethod


class InputDataToSurfexBinaries(ABC):
    """Abstract input data."""

    @abstractmethod
    def __init__(self):
        """Construct."""
        raise NotImplementedError

    @abstractmethod
    def prepare_input(self):
        """Prepare input."""
        raise NotImplementedError


class OutputDataFromSurfexBinaries(ABC):
    """Abstract output data."""

    @abstractmethod
    def __init__(self):
        """Construct."""
        raise NotImplementedError

    @abstractmethod
    def archive_files(self):
        """Archive files."""
        raise NotImplementedError


class JsonOutputData(OutputDataFromSurfexBinaries):
    """Output data."""

    def __init__(self, data):
        """Output data from dict.

        Args:
            data (dict): Output data.

        """
        self.data = data

    def archive_files(self):
        """Archive files."""
        for output_file, ltarget in self.data.items():
            target = ltarget
            logging.info("%s -> %s", output_file, target)
            command = "mv"
            if isinstance(target, dict):
                for key in target:
                    logging.debug("%s %s %s", output_file, key, target[key])
                    command = target[key]
                    target = key

            cmd = command + " " + output_file + " " + target
            try:
                logging.info(cmd)
                subprocess.check_call(cmd, shell=True)  # noqa S602
            except IOError:
                logging.error("%s failed", cmd)
                raise RuntimeError(cmd + " failed") from IOError


class JsonOutputDataFromFile(JsonOutputData):
    """JSON output data."""

    def __init__(self, file):
        """Construct from json file."""
        with open(file, mode="r", encoding="utf-8") as file_handler:
            data = json.load(file_handler)
        JsonOutputData.__init__(self, data)

    def archive_files(self):
        """Archive files."""
        JsonOutputData.archive_files(self)


class JsonInputData(InputDataToSurfexBinaries):
    """JSON input data."""

    def __init__(self, data):
        """Construct input data.

        Args:
            data (dict): Input data.
        """
        self.data = data

    def prepare_input(self):
        """Prepare input."""
        for target, linput_file in self.data.items():
            input_file = linput_file
            logging.info("%s -> %s", target, input_file)
            logging.debug(os.path.realpath(target))
            command = None
            if isinstance(input_file, dict):
                for lkey in input_file:
                    key = lkey
                    logging.debug(key)
                    logging.debug(input_file[key])
                    command = str(input_file[key])
                    input_file = str(key)
                    command = command.replace("@INPUT@", input_file)
                    command = command.replace("@TARGET@", target)

            if os.path.realpath(target) == os.path.realpath(input_file):
                logging.info("Target and input file is the same file")
            else:
                if command is None:
                    cmd = "ln -sf " + input_file + " " + target
                else:
                    cmd = command
                try:
                    logging.info(cmd)
                    subprocess.check_call(cmd, shell=True)  # noqa S602
                except IOError:
                    raise (cmd + " failed") from IOError

    def add_data(self, data):
        """Add data.

        Args:
            data (dict): Data to add
        """
        for key in data:
            value = data[key]
            self.data.update({key: value})


class JsonInputDataFromFile(JsonInputData):
    """JSON input data."""

    def __init__(self, file):
        """Construct JSON input data.

        Args:
            file (str): JSON file name

        """
        with open(file, mode="r", encoding="utf-8") as file_handler:
            data = json.load(file_handler)
        JsonInputData.__init__(self, data)

    def prepare_input(self):
        """Prepare input."""
        JsonInputData.prepare_input(self)


class InputDataFromNamelist(JsonInputData):
    """Binary input data for offline executables."""

    def __init__(
        self,
        nml,
        input_data,
        program,
        platform,
        basetime=None,
        validtime=None,
    ):
        """Construct InputDataFromNamelist.

        Args:
            nml (f90nml.Namelist): Namelist
            input_data (dict): Input data mapping
            program (str): Kind of program
            platform (SystemFilePaths): Platform settings
            basetime (as_datetime, optional): Baseetime. Defaults to None.
            validtime (as_datetime, optional): Validtime. Defaults to None.

        Raises:
            RuntimeError: Program not defined

        """
        self.nml = nml
        self.platform = platform
        self.basetime = basetime
        self.validtime = validtime
        try:
            self.data = input_data[program]
        except KeyError:
            raise RuntimeError(f"Could not find program {program}") from KeyError
        data = self.process_data()
        JsonInputData.__init__(self, data)

    @staticmethod
    def get_nml_value2(nml, block, key, indices=None):
        """Get namelist value.

        Args:
            nml (nmlf90.Namelist): Namelist
            block (str): Namelist block
            key (str): Namelist key
            indices (list, optional): Indices to read. Defaults to None.

        Returns:
            setting (any): Namelist setting

        """
        logging.debug("Checking block=%s key=%s", block, key)
        if block in nml and key in nml[block]:
            logging.debug(nml[block][key])
            if indices is not None:
                logging.debug("indices=%s", indices)
                try:
                    if len(indices) == 2:
                        val = nml[block][key][indices[1]][indices[0]]
                    else:
                        val = nml[block][key][indices[0]]
                        logging.debug("Found 1D value %s", val)
                        if isinstance(val, list):
                            return None
                except IndexError:
                    return None
                except TypeError:
                    return None
            else:
                val = nml[block][key]

            logging.debug("Found: %s Indices=%s", val, indices)
            return val
        return None

    @staticmethod
    def get_nml_value(nml, block, key, indices=None):
        """Get namelist value.

        Args:
            nml (nmlf90.Namelist): Namelist
            block (str): Namelist block
            key (str): Namelist key
            indices (list, optional): Indices to read. Defaults to None.

        Returns:
            setting (any): Namelist setting

        """
        logging.debug("Checking block=%s key=%s", block, key)
        if block in nml and key in nml[block]:
            vals = []
            val_dict = {}
            val = nml[block][key]
            logging.debug("namelist type=%s", type(val))
            if indices is not None:
                logging.debug("indices=%s", indices)
                if len(indices) == 2:
                    val = nml[block][key][indices[1]][indices[0]]
                else:
                    val = nml[block][key][indices[0]]
                    logging.debug("Found 1D value %s", val)
                    if isinstance(val, list):
                        return None
                val_dict.update({"value": val, "indices": None})
                vals.append(val_dict)

            elif isinstance(val, list):
                dim_size = len(val)
                logging.debug("dim_size=%s", dim_size)
                dims = []
                tval = val
                more_dimensions = True
                while more_dimensions:
                    logging.debug("tval=%s type(tval)=%s", tval, type(tval))
                    if (
                        isinstance(tval, int)
                        or not isinstance(tval, list)
                        or isinstance(tval[0], int)
                    ):
                        more_dimensions = False
                    else:
                        logging.debug("len(tval)=%s type(tval)=%s", len(tval), type)
                        dim_size = len(tval)
                        dims.append(dim_size)

                        tval = tval[0]
                        logging.debug("New tval=%s dim_size=%s", tval, dim_size)

                logging.debug("dims=%s", dims)
                logging.debug("type(val)=%s", type(val))
                if len(dims) == 2:
                    for i in range(dims[0]):
                        for j in range(dims[1]):
                            val_dict = {}
                            indices = [j, i]
                            lval = val[i][j]
                            val_dict.update({"value": lval, "indices": indices})
                            logging.debug("value=%s indices=%s", lval, indices)
                            vals.append(val_dict)
                elif len(dims) == 1:
                    for i in range(dims[0]):
                        val_dict = {}
                        indices = [i]
                        logging.debug("i=%s, val[i]=%s", i, val[i])
                        lval = val[i]
                        val_dict.update({"value": lval, "indices": indices})
                        logging.debug("value=%s indices=%s", lval, indices)
                        vals.append(val_dict)
                elif len(dims) == 0:
                    val_dict = {}
                    logging.debug("val=%s", val)
                    val_dict.update({"value": val, "indices": None})
                    vals.append(val_dict)
            else:
                val_dict = {}
                if isinstance(val, bool):
                    val = str(val)
                val_dict.update({"value": val, "indices": None})
                vals.append(val_dict)

            logging.debug("Found: value=%s", val_dict["value"])
            return vals
        return None

    @staticmethod
    def get_nml_value_from_string(nml, string, sep="#", indices=None):
        """Get namelist value from a string.

        Args:
            nml (nmlf90.Namelist): Namelist
            string (str): Namelist identifier
            sep (str, optional): _description_. Defaults to "#".
            indices (list, optional): Indices to read. Defaults to None.

        Returns:
            setting (any): Namelist setting

        """
        nam_section = string.split(sep)[0]
        nam_key = string.split(sep)[1]
        return InputDataFromNamelist.get_nml_value(
            nml, nam_section, nam_key, indices=indices
        )

    def substitute(self, key, val, macros=None, micro="@", check_parsing=False):
        """Substitute patterns.

        Args:
            key (str): _description_
            val (str): _description_
            macros (dict, optional): Macros. Defaults to None.
            micro (str, optional): Micro character. Defaults to "@".
            check_parsing (bool, optional): Caheck if values were substituted.

        Returns:
            dict: Substituted key=value

        """
        logging.debug(
            "Substitute key=%s and val=%s %s %s", key, val, self.basetime, self.validtime
        )
        pkey = key
        pval = val
        for spath_key, spath_val in self.platform.system_file_paths.items():
            if isinstance(spath_val, str):
                pkey = pkey.replace(f"{micro}{spath_key}{micro}", spath_val)
                pval = pval.replace(f"{micro}{spath_key}{micro}", spath_val)
        if macros is not None:
            for macro_key, macro_val in macros.items():
                pkey = pkey.replace(f"{micro}{macro_key}{micro}", macro_val)
                pval = pval.replace(f"{micro}{macro_key}{micro}", macro_val)

        pkey = self.platform.parse_setting(
            pkey,
            validtime=self.validtime,
            basedtg=self.basetime,
            check_parsing=check_parsing,
        )
        pval = self.platform.parse_setting(
            pval,
            validtime=self.validtime,
            basedtg=self.basetime,
            check_parsing=check_parsing,
        )
        return pkey, pval

    def read_macro_setting(self, macro_defs, key, default=None, sep="#"):
        """Read a macro setting.

        Args:
            macro_defs (dict): Macro definition
            key (str): Macro setting to get.
            default (str, optional): Default value. Defaults to None.
            sep (str, optional): Namelist key separator. Defaults to "#".

        Returns:
            setting (any)
        """
        try:
            setting = macro_defs[key]
            if isinstance(setting, str) and setting.find(sep) > 0:
                logging.debug("Read macro setting from namelist %s", setting)
                setting = self.get_nml_value_from_string(self.nml, setting)
                if isinstance(setting, list):
                    setting = setting[0]["value"]
            return setting
        except KeyError:
            return default

    def extend_macro(self, key, val, macros, sep="#"):
        """Extend entries from macro.

        Args:
            key (_type_): _description_
            val (_type_): _description_
            macros (dict): Macros
            sep (str, optional): Namelist key separator. Defaults to "#".

        Raises:
            NotImplementedError: _description_
            NotImplementedError: _description_

        Returns:
            dict: Key, value dictionary
        """
        logging.debug("extenders=%s", macros)
        if macros is None:
            return {key: val}

        processed_data = {}
        for macro, macro_types in macros.items():
            loop = {}
            for macro_type, macro_defs in macro_types.items():
                logging.debug("macro_defs=%s", macro_defs)

                if macro_type == "ekfpert":
                    nncvs = self.read_macro_setting(macro_defs, "list", sep=sep)
                    logging.debug("nncvs=%s", nncvs)
                    nncvs = nncvs.copy()
                    duplicate = self.read_macro_setting(macro_defs, "duplicate", sep=sep)
                    if duplicate:
                        nncvs += nncvs
                    loop.update({"0": "0"})
                    icounter1 = 1
                    icounter2 = 1
                    for nncv in nncvs:
                        if nncv == 1:
                            loop.update({str(icounter1): str(icounter2)})
                            icounter1 += 1
                        icounter2 += 1
                elif macro_type == "dict":
                    values = self.get_nml_value_from_string(self.nml, macro_defs)
                    counter = 0
                    for lkey, lval in values[0].items():
                        loop.update({str(lkey): str(lval)})
                        counter += 1  # noqa SIM113

                elif macro_type == "iterator":
                    start = self.read_macro_setting(macro_defs, "start", sep=sep)
                    end = self.read_macro_setting(macro_defs, "end", sep=sep)
                    fmt = self.read_macro_setting(
                        macro_defs, "fmt", sep=sep, default=None
                    )
                    if fmt is None:
                        fmt = "{:d}"
                    for lval in range(start, end):
                        loop.update({str(fmt.format(lval)): str(fmt.format(lval))})
                else:
                    raise NotImplementedError

            # Loop normal macros not being nml arrays
            unprocessed_data = processed_data.copy()
            unprocessed_data = processed_data if processed_data else {key: val}

            for lkey, lval in unprocessed_data.items():
                mkey = lkey
                for vmacro1, vmacro2 in loop.items():
                    logging.debug(
                        "key=%s val=%s macro=%s vmacro1=%s vmacro2=%s",
                        mkey,
                        lval,
                        macro,
                        vmacro1,
                        vmacro2,
                    )
                    if mkey.find("#") > 0:
                        mkey = self.get_nml_value_from_string(self.nml, mkey, sep=sep)
                    pkey = mkey.replace(f"@{macro}@", vmacro1)
                    pval = lval.replace(f"@{macro}@", vmacro2)
                    processed_data.update({pkey: pval})

        logging.debug("Processed data=%s", processed_data)
        return processed_data

    def process_macro(self, key, val, macros, sep="#", indices=None):
        """Process macro.

        Args:
            key (str): Key
            val (str): Value
            macros (dict): Macros
            sep (str, optional): Namelist key separator. Defaults to "#".
            indices (list, optional): Process macro from namelist indices.

        Raises:
            NotImplementedError: Only 2 dimensions are implemented

        Returns:
            dict: Key, value dictionary
        """
        logging.debug("macros=%s", macros)
        if macros is None:
            return key, val

        logging.debug("indices=%s", indices)
        if indices is None:
            return key, val

        pkey = key
        pval = val
        for macro in macros:
            lindex = None
            if len(indices) == 2:
                lindex = indices[1] if macro == "DECADE" else indices[0]
            elif len(indices) == 1:
                lindex = indices[0]
            elif len(indices) > 2:
                raise NotImplementedError("Only 2 dimensions are implemented")

            vmacro = None
            if lindex is not None:
                try:
                    macro_defs = macros[macro]
                    logging.debug("macro_defs=%s", macro_defs)
                except KeyError:
                    logging.warning(
                        "Macro %s not defined. Use index value %s", macro, lindex
                    )
                    vmacro = str(lindex + 1)

                if macro == "VTYPE":
                    vmacro = str(lindex + 1)
                elif "DECADE" in macros:
                    ntime = self.read_macro_setting(macro_defs, "ntime", sep=sep)
                    dec_days = int(360 / float(ntime))
                    dec_start = int(dec_days / 2)
                    dec_end = 360 + dec_start
                    dec = 0
                    for day in range(dec_start, dec_end, dec_days):
                        logging.debug("day=%s, dec=%s lindex=%s", day, dec, lindex)
                        month = int(day / 30) + 1
                        mday = int(day % 30)
                        if dec == lindex:
                            vmacro = f"{month:02d}{mday:02d}"
                        dec += 1  # noqa SIM113

                logging.debug(
                    "Substitute @%s@ with %s pkey=%s pval=%s", macro, vmacro, pkey, pval
                )
                if isinstance(pkey, str):
                    pkey = pkey.replace(f"@{macro}@", vmacro)
                if isinstance(pval, str):
                    pval = pval.replace(f"@{macro}@", vmacro)
                logging.debug(
                    "Substitute @%s@ with %s pkey=%s pval=%s", macro, vmacro, pkey, pval
                )
        return pkey, pval

    def matching_value(self, data, val, sep="#", indices=None):
        """Match the value. Possibly also read namelist value.

        Args:
            data (dict): Data to check keys for
            val (str): Key to find
            sep (str, optional): Namelist separator. Defaults to "#".
            indices(list, optional): Indices in namelist

        Raises:
            RuntimeError: "Malformed input data"

        Returns:
            dict: Matching entry in data.

        """
        if val in ("macro", "extenders"):
            return None
        logging.debug("type(data)=%s", type(data))
        logging.debug("type(val)=%s", type(val))
        logging.debug("indices=%s", indices)
        mdata = data.keys() if isinstance(data, dict) else [data]
        val = str(val)
        logging.debug("Check if val=%s matches mdata=%s", val, mdata)
        sval = None
        for mval in mdata:
            if val.find(sep) > 0:
                logging.debug("val=%s is a namelist variable", val)
                sval = self.get_nml_value_from_string(self.nml, val, indices=indices)
                logging.debug("Got sval=%s", sval)
                if sval is None:
                    return None
                indices = sval[0]["indices"]
                sval = sval[0]["value"]
            if mval == val:
                logging.debug("Found matching data. val=%s data=%s", val, data)
                try:
                    rval = data[val]
                except TypeError:
                    raise RuntimeError("Malformed input data") from TypeError
                if sval is not None:
                    rval = {sval: rval}
                logging.debug("Return data rval=%s", rval)
                return rval
        logging.warning("Value=%s not found in data", val)
        return None

    def process_data(self, sep="#"):
        """Process input definitions on files to map.

        Args:
            sep (str, optional): Namelist separator. Defaults to "#".

        Returns:
            mapped_data (dict): A dict with mapped local names and target files.

        """
        logging.debug("Process data: %s", self.data)

        def _process_data(mapped_data, data, indices=None, macros=None, extenders=None):
            for key, value in data.items():
                logging.debug(".................. key=%s", key)
                # Required namelist variable
                if key.find(sep) > 0:
                    vals = self.get_nml_value_from_string(self.nml, key, indices=indices)
                else:
                    vals = [{"value": value, "indices": None}]

                if isinstance(vals, list):
                    for val_dict in vals:
                        logging.debug("=========== val_dict=%s", val_dict)
                        val = val_dict["value"]
                        indices = val_dict["indices"]

                        setting = self.matching_value(
                            value, val, sep=sep, indices=indices
                        )
                        logging.debug("Setting=%s", setting)
                        if setting is not None:
                            if "macros" in setting:
                                macros = setting.copy()
                                macros = macros["macros"]
                            if "extenders" in setting:
                                extenders = setting.copy()
                                extenders = extenders["extenders"]

                            last_dict = True
                            if isinstance(setting, dict):
                                for tval in setting.values():
                                    if isinstance(tval, dict):
                                        last_dict = False
                            else:
                                logging.debug("setting=%s", setting)
                            if not last_dict:
                                logging.debug(
                                    "------ Call next loop. setting=%s", setting
                                )
                                _process_data(
                                    mapped_data,
                                    setting,
                                    indices=indices,
                                    macros=macros,
                                    extenders=extenders,
                                )
                            else:
                                for lkey2, value2 in setting.items():
                                    key2 = lkey2
                                    logging.debug(
                                        "Setting1 key=%s value=%s indices=%s",
                                        key2,
                                        value2,
                                        indices,
                                    )
                                    if key2.find(sep) > 0:
                                        keys = self.get_nml_value_from_string(
                                            self.nml, key2, indices=indices
                                        )
                                        key2 = keys[0]["value"]

                                    processed = False
                                    logging.debug(
                                        "Setting2 key=%s value=%s indices=%s",
                                        key2,
                                        value2,
                                        indices,
                                    )
                                    if macros is not None:
                                        processed = True
                                        key3, value3 = self.process_macro(
                                            key2, value2, macros, indices=indices
                                        )
                                        if value3.endswith(".dir"):
                                            dir_key = key3 + ".dir"
                                            dir_val = value3
                                            hdr_key = key3 + ".hdr"
                                            hdr_val = value3.replace(".dir", ".hdr")
                                            hdr_key, hdr_val = self.substitute(
                                                hdr_key, hdr_val
                                            )
                                            dir_key, dir_val = self.substitute(
                                                dir_key, dir_val
                                            )
                                            mapped_data.update({hdr_key: hdr_val})
                                            mapped_data.update({dir_key: dir_val})
                                        elif value3.endswith(".nc"):
                                            my_key, my_val = self.substitute(key3, value3)
                                            logging.debug(
                                                "my_key=%s, my_val=%s", my_key, my_val
                                            )
                                            if not my_key.endswith(".nc"):
                                                my_key = my_key + ".nc"
                                            mapped_data.update({my_key: my_val})
                                        else:
                                            my_key, my_val = self.substitute(key3, value3)
                                            mapped_data.update({my_key: my_val})

                                    if extenders is not None:
                                        processed = True
                                        processed_values = self.extend_macro(
                                            key2, value2, extenders
                                        )
                                        for pkey3, pval3 in processed_values.items():
                                            logging.debug(
                                                "pkey3=%s pval3=%s", pkey3, pval3
                                            )
                                            mpkey3, mpval3 = self.substitute(pkey3, pval3)
                                            logging.debug(
                                                "mpkey3=%s mpval3=%s", mpkey3, mpval3
                                            )
                                            mapped_data.update({mpkey3: mpval3})

                                    if not processed:
                                        pkey3 = key2
                                        pval3 = value2
                                        logging.debug("pkey3=%s pval3=%s", pkey3, pval3)
                                        if pval3.endswith(".nc") and not pkey3.endswith(
                                            ".nc"
                                        ):
                                            pkey3 = pkey3 + ".nc"
                                        pkey3, pval3 = self.substitute(pkey3, pval3)
                                        mapped_data.update({pkey3: pval3})

                                indices = None
                        elif key not in ["macros", "extenders"]:
                            logging.warning("Could not match key=%s value=%s", key, val)
                else:
                    logging.warning("Could not find namelist key=%s", key)
                    indices = None

        mapped_data = {}
        _process_data(mapped_data, self.data)
        logging.debug("Mapped data=%s", mapped_data)
        return mapped_data



"""Forcing."""

import abc
import copy
import json
import logging
import os
import shutil
import time

try:
    import netCDF4
except ModuleNotFoundError:
    logging.debug("netCDF4 not loaded")
import numpy as np

try:
    import yaml
except ModuleNotFoundError:
    logging.debug("yaml not loaded")

from . import PACKAGE_DIRECTORY
from .cache import Cache
from .datetime_utils import as_datetime, as_timedelta
from .file import ForcingFileNetCDF
from .geo import get_geo_object
from .read import ConstantValue, ConvertedInput, Converter
from .util import deep_update


# TODO: should be abstract?
class SurfexForcing(object):
    """Surfex forcing class."""

    def __init__(self):
        """Construct object."""


class SurfexNetCDFForcing(SurfexForcing):
    """NetCDF surfex forcing."""

    def __init__(self, filename, geo):
        """Construct netcdf forcing."""
        SurfexForcing.__init__(self)
        self.io_object = ForcingFileNetCDF(filename, geo)


class SurfexOutputForcing(object):
    """Main output class for SURFEX forcing."""

    __metaclass__ = abc.ABCMeta

    def __init__(self, base_time, geo, ntimes, var_objs, time_step_intervall):
        """Construct output forcing.

        Args:
            base_time (_type_): _description_
            geo (_type_): _description_
            ntimes (_type_): _description_
            var_objs (_type_): _description_
            time_step_intervall (_type_): _description_
        """
        self.time_step_intervall = time_step_intervall
        self.valid_time = None
        self.base_time = base_time
        self.geo = geo
        self.ntimes = ntimes
        self.time_step = 0
        self.time_step_value = 0
        self.var_objs = var_objs
        self._check_sanity()

    def _check_sanity(self):
        if len(self.var_objs) != self.nparameters:
            raise RuntimeError(
                f"Inconsistent number of parameter. {len(self.var_objs)!s} != "
                f"{self.nparameters!s}"
            )

        # Check if all parameters are present
        for var_obj in self.var_objs:
            self.parameters[var_obj.var_name] = 1

        for key, value in self.parameters.items():
            if value == 0:
                raise RuntimeError(f"Required parameter {key!s} is missing!")

    # Time dependent parameter
    nparameters = 11
    parameters = {
        "TA": 0,
        "QA": 0,
        "PS": 0,
        "DIR_SW": 0,
        "SCA_SW": 0,
        "LW": 0,
        "RAIN": 0,
        "SNOW": 0,
        "WIND": 0,
        "WIND_DIR": 0,
        "CO2": 0,
    }

    @abc.abstractmethod
    def write_forcing(self, var_objs, this_time, cache):
        """Write forcing."""
        raise NotImplementedError("users must define writeForcing to use this base class")


class NetCDFOutput(SurfexOutputForcing):
    """Forcing in NetCDF format."""

    forcing_file = {}

    translation = {
        "TA": "Tair",
        "QA": "Qair",
        "PS": "PSurf",
        "DIR_SW": "DIR_SWdown",
        "SCA_SW": "SCA_SWdown",
        "LW": "LWdown",
        "RAIN": "Rainf",
        "SNOW": "Snowf",
        "WIND": "Wind",
        "WIND_DIR": "Wind_DIR",
        "CO2": "CO2air",
    }

    def __init__(
        self,
        base_time,
        geo,
        fname,
        ntimes,
        var_objs,
        att_objs,
        att_time,
        cache,
        time_step,
        fmt="netcdf",
        diskless_write=False,
    ):
        """Construct netcdf forcing.

        Args:
            base_time (_type_): _description_
            geo (_type_): _description_
            fname (_type_): _description_
            ntimes (_type_): _description_
            var_objs (_type_): _description_
            att_objs (_type_): _description_
            att_time (_type_): _description_
            cache (_type_): _description_
            time_step (_type_): _description_
            fmt (str, optional): _description_. Defaults to "netcdf".
            diskless_write: sets diskelss=True and persist=True

        Raises:
            NotImplementedError: NotImplementedError

        """
        SurfexOutputForcing.__init__(self, base_time, geo, ntimes, var_objs, time_step)
        if fmt == "netcdf":
            self.output_format = "NETCDF3_64BIT"
        elif fmt == "nc4":
            self.output_format = "NETCDF4"
        else:
            raise NotImplementedError(format)
        logging.info("Forcing type is netCDF")
        self.forcing_file = {}
        if fname is None:
            fname = "FORCING.nc"
        self.fname = fname
        self.tmp_fname = self.fname + ".tmp"
        self.file_handler = netCDF4.Dataset(
            self.tmp_fname,
            "w",
            format=self.output_format,
            diskless=diskless_write,
            persist=diskless_write,
        )
        self._define_forcing(geo, att_objs, att_time, cache)

    def write_forcing(self, var_objs, this_time, cache):  # noqa ARG002
        """Write forcing.

        Args:
            var_objs (_type_): _description_
            this_time (_type_): _description_
            cache (_type_): _description_

        """
        # VARS
        for this_obj in self.var_objs:
            this_var = this_obj.var_name
            logging.info("Preparing variable %s", this_obj.var_name)
            tic = time.time()
            field = this_obj.read_time_step(this_time, cache)
            field = field.reshape([self.geo.nlats, self.geo.nlons], order="F").flatten()
            toc = time.time()
            logging.info("Preparation took %s seconds", str(toc - tic))
            self.forcing_file[self.translation[this_var]][self.time_step, :] = field

        self.forcing_file["TIME"][self.time_step] = self.time_step_value

    def _define_forcing(self, geo, att_objs, att_time, cache):
        logging.info("Define netcdf forcing")

        zs_oro = None
        zref = None
        uref = None
        for this_obj in att_objs:
            this_var = this_obj.var_name
            logging.info("Define: %s", this_obj.var_name)
            if this_var == "ZS":
                zs_oro = this_obj.read_time_step(att_time, cache)
                zs_oro = zs_oro.reshape(
                    [self.geo.nlats, self.geo.nlons], order="F"
                ).flatten()
            elif this_var == "ZREF":
                zref = this_obj.read_time_step(att_time, cache)
                zref = zref.reshape([self.geo.nlats, self.geo.nlons], order="F").flatten()
            elif this_var == "UREF":
                uref = this_obj.read_time_step(att_time, cache)
                uref = uref.reshape([self.geo.nlats, self.geo.nlons], order="F").flatten()

        # DIMS
        self.forcing_file["NPOINTS"] = self.file_handler.createDimension(
            "Number_of_points", geo.npoints
        )
        self.forcing_file["NTIMES"] = self.file_handler.createDimension(
            "time", self.ntimes
        )

        # DEFINE VARS
        self.forcing_file["TIME"] = self.file_handler.createVariable(
            "time", "f4", ("time",)
        )
        self.forcing_file["TIME"].units = (
            "hours since " + f"{self.base_time.strftime('%Y-%m-%d %H')}:00:00 0:00"
        )
        self.forcing_file["TSTEP"] = self.file_handler.createVariable(
            "FRC_TIME_STP", "f4"
        )
        self.forcing_file["TSTEP"].longname = "Forcing_Time_Step"
        self.forcing_file["TSTEP"][:] = self.time_step_intervall
        self.forcing_file["LON"] = self.file_handler.createVariable(
            "LON", "f4", ("Number_of_points",)
        )
        self.forcing_file["LON"].longname = "Longitude"
        self.forcing_file["LON"][:] = geo.lons.flatten(order="F")
        self.forcing_file["LAT"] = self.file_handler.createVariable(
            "LAT", "f4", ("Number_of_points",)
        )
        self.forcing_file["LAT"].longname = "Latitude"
        self.forcing_file["LAT"][:] = geo.lats.flatten(order="F")
        self.forcing_file["ZS"] = self.file_handler.createVariable(
            "ZS", "f4", ("Number_of_points",)
        )
        self.forcing_file["ZS"].longname = "Surface_Orography"
        self.forcing_file["ZS"][:] = zs_oro
        self.forcing_file["ZREF"] = self.file_handler.createVariable(
            "ZREF", "f4", ("Number_of_points",)
        )
        self.forcing_file["ZREF"].longname = "Reference_Height"
        self.forcing_file["ZREF"].units = "m"
        self.forcing_file["ZREF"][:] = zref
        self.forcing_file["UREF"] = self.file_handler.createVariable(
            "UREF", "f4", ("Number_of_points",)
        )
        self.forcing_file["UREF"].longname = "Reference_Height_for_Wind"
        self.forcing_file["UREF"].units = "m"
        self.forcing_file["UREF"][:] = uref

        # Define time dependent variables
        for this_obj in self.var_objs:
            this_var = this_obj.var_name

            if this_var == "TA":
                self.forcing_file["Tair"] = self.file_handler.createVariable(
                    "Tair",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["Tair"].longname = "Near_Surface_Air_Temperature"
                self.forcing_file["Tair"].units = "K"
            elif this_var == "QA":
                self.forcing_file["Qair"] = self.file_handler.createVariable(
                    "Qair",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["Qair"].longname = "Near_Surface_Specific_Humidity"
                self.forcing_file["Qair"].units = "kg/kg"
            elif this_var == "PS":
                self.forcing_file["PSurf"] = self.file_handler.createVariable(
                    "PSurf",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["PSurf"].longname = "Surface_Pressure"
                self.forcing_file["PSurf"].units = "Pa"
            elif this_var == "DIR_SW":
                self.forcing_file["DIR_SWdown"] = self.file_handler.createVariable(
                    "DIR_SWdown",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["DIR_SWdown"].longname = (
                    "Surface_Incident_Downwelling_Shortwave_Radiation"
                )
                self.forcing_file["DIR_SWdown"].units = "W/m2"
            elif this_var == "SCA_SW":
                self.forcing_file["SCA_SWdown"] = self.file_handler.createVariable(
                    "SCA_SWdown",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["SCA_SWdown"].longname = (
                    "Surface_Incident_Diffuse_Shortwave_Radiation"
                )
                self.forcing_file["SCA_SWdown"].units = "W/m2"
            elif this_var == "LW":
                self.forcing_file["LWdown"] = self.file_handler.createVariable(
                    "LWdown",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["LWdown"].longname = (
                    "Surface_Incident_Diffuse_Longwave_Radiation"
                )
                self.forcing_file["LWdown"].units = "W/m2"
            elif this_var == "RAIN":
                self.forcing_file["Rainf"] = self.file_handler.createVariable(
                    "Rainf",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["Rainf"].longname = "Rainfall_Rate"
                self.forcing_file["Rainf"].units = "kg/m2/s"
            elif this_var == "SNOW":
                self.forcing_file["Snowf"] = self.file_handler.createVariable(
                    "Snowf",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["Snowf"].longname = "Snowfall_Rate"
                self.forcing_file["Snowf"].units = "kg/m2/s"
            elif this_var == "WIND":
                self.forcing_file["Wind"] = self.file_handler.createVariable(
                    "Wind",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["Wind"].longname = "Wind_Speed"
                self.forcing_file["Wind"].units = "m/s"
            elif this_var == "WIND_DIR":
                self.forcing_file["Wind_DIR"] = self.file_handler.createVariable(
                    "Wind_DIR",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["Wind_DIR"].longname = "Wind_Direction"
            elif this_var == "CO2":
                self.forcing_file["CO2air"] = self.file_handler.createVariable(
                    "CO2air",
                    "f4",
                    (
                        "time",
                        "Number_of_points",
                    ),
                )
                self.forcing_file["CO2air"].longname = "Near_Surface_CO2_Concentration"
                self.forcing_file["CO2air"].units = "kg/m3"
            else:
                raise NotImplementedError(
                    f"This should never happen! {this_var} is not defined!"
                )

    def finalize(self):
        """Finalize the forcing. Close the file."""
        logging.debug("Close file")
        self.file_handler.close()
        shutil.move(self.tmp_fname, self.fname)


class AsciiOutput(SurfexOutputForcing):
    """Forcing in ASCII format."""

    def __init__(
        self,
        base_time,
        geo,
        fname,
        ntimes,
        var_objs,
        att_objs,
        att_time,
        cache,
        time_step,
    ):
        """Construct ASCII forcing output."""
        SurfexOutputForcing.__init__(self, base_time, geo, ntimes, var_objs, time_step)
        self.output_format = "ascii"
        logging.info("Forcing type is ASCII")
        self.forcing_file = {}
        self.file_handler = {}
        if fname is None:
            fname = "Params_config.txt"
        self.fname = fname
        self._define_forcing(geo, att_objs, att_time, cache)

    def write_forcing(self, var_objs, this_time, cache):  # noqa ARG002
        """Write forcing.

        Args:
            var_objs (_type_): _description_
            this_time (_type_): _description_
            cache (_type_): _description_

        """
        for this_obj in self.var_objs:
            this_var = this_obj.var_name
            logging.info("Write var name %s", this_obj.var_name)
            field = this_obj.read_time_step(this_time, cache)
            fmt = "%20.8f"
            cols = 50
            write_formatted_array(self.file_handler[this_var], field, cols, fmt)

    def _define_forcing(self, geo, att_objs, att_time, cache):
        zs_oro = None
        zref = None
        uref = None
        for this_obj in att_objs:
            this_var = this_obj.var_name
            if this_var == "ZS":
                zs_oro = this_obj.read_time_step(att_time, cache)
            elif this_var == "ZREF":
                zref = this_obj.read_time_step(att_time, cache)
            elif this_var == "UREF":
                uref = this_obj.read_time_step(att_time, cache)

        second = (
            self.base_time
            - self.base_time.replace(hour=0, minute=0, second=0, microsecond=0)
        ).total_seconds()
        fmt = "%15.8f"
        cols = 50
        with open(self.fname, mode="w", encoding="utf-8") as file_handler:
            file_handler.write(str(geo.npoints) + "\n")
            file_handler.write(str(self.ntimes) + "\n")
            file_handler.write(str(self.time_step_intervall) + "\n")
            file_handler.write(self.base_time.strftime("%Y") + "\n")
            file_handler.write(self.base_time.strftime("%m") + "\n")
            file_handler.write(self.base_time.strftime("%d") + "\n")
            file_handler.write(str(second) + "\n")
            write_formatted_array(file_handler, geo.lons, cols, fmt)
            write_formatted_array(file_handler, geo.lats, cols, fmt)
            write_formatted_array(file_handler, zs_oro, cols, fmt)
            write_formatted_array(file_handler, zref, cols, fmt)
            write_formatted_array(file_handler, uref, cols, fmt)

        for key in self.parameters:
            nam = key
            if key == "WIND_DIR":
                nam = "DIR"
            self.forcing_file[key] = "Forc_" + nam + ".txt"
            fhandler = open(self.forcing_file[key], mode="w", encoding="utf-8")  # noqa
            self.file_handler[key] = fhandler

    def finalize(self):
        """Finalize forcing."""
        logging.debug("Close file")
        for key in self.parameters:
            self.file_handler[key].close()


def write_formatted_array(file, array, columns, fileformat):
    """Write a formatted array."""
    full_lines = int(array.size / columns)
    extra = array.size % columns
    first = array.size - extra
    astr = np.empty(columns * full_lines, dtype="float64")
    array = array.flatten()
    astr = array[0 : columns * full_lines]
    astr1 = astr.reshape(columns, full_lines, order="F")
    np.savetxt(file, astr1.T, fmt=fileformat, newline="\n", delimiter="")
    if extra != 0:
        astr2 = np.empty(extra, dtype="float64")
        astr2 = array[first : array.size]
        astr2 = astr2.reshape(extra, 1, order="F")
        np.savetxt(file, astr2.T, fmt=fileformat, newline="\n", delimiter="")


def run_time_loop(options, var_objs, att_objs):
    """Run time loop."""
    tic = time.time()
    this_time = options["start"]

    cache = Cache(options["cache_interval"])
    time_step = options["timestep"]
    single = False
    if "single" in options:
        single = options["single"]

    # Find how many time steps we want to write
    ntimes = 0
    while this_time <= options["stop"]:
        ntimes = ntimes + 1
        this_time = this_time + as_timedelta(seconds=options["timestep"])
    if single:
        time_step = 1
        if ntimes == 1:
            ntimes = 2
            logging.info("Print single time step twice %s", 0)
        else:
            raise RuntimeError("Option single should be used with one time step")

    output_file = options["output_file"]
    if os.path.exists(output_file) and not options["force"]:
        logging.info("%s already exists and force in false", output_file)
        return

    if os.path.exists(output_file):
        logging.warning("Overwrite output: %s", output_file)

    # Create output object
    if (
        str.lower(options["output_format"]) == "netcdf"
        or str.lower(options["output_format"]) == "nc4"
    ):
        # Set att_time the same as start
        att_time = options["start"]
        output = NetCDFOutput(
            options["start"],
            options["geo_out"],
            options["output_file"],
            ntimes,
            var_objs,
            att_objs,
            att_time,
            cache,
            time_step,
            fmt=str.lower(options["output_format"]),
            diskless_write=options["diskless_write"],
        )
    elif str.lower(options["output_format"]) == "ascii":
        att_time = options["start"]
        output = AsciiOutput(
            options["start"],
            options["geo_out"],
            options["output_file"],
            ntimes,
            var_objs,
            att_objs,
            att_time,
            cache,
            time_step,
        )
    else:
        raise NotImplementedError("Invalid output format " + options["output_format"])

    # Loop output time steps
    this_time = options["start"]
    while this_time <= options["stop"]:
        # Write for each time step
        logging.info(
            "Creating forcing for: %s  time_step: %s",
            this_time.strftime("%Y%m%d%H"),
            str(output.time_step),
        )
        output.write_forcing(var_objs, this_time, cache)
        output.time_step = output.time_step + 1
        if not single:
            output.time_step_value = output.time_step
            this_time = this_time + as_timedelta(seconds=options["timestep"])
            if cache is not None:
                cache.clean_fields(this_time)
        else:
            output.time_step_value = 0
            if output.time_step > 1:
                this_time = this_time + as_timedelta(seconds=options["timestep"])

    # Finalize forcing
    output.finalize()
    toc = time.time()
    logging.info("Forcing generation took %s seconds", str(toc - tic))


def set_input_object(
    sfx_var,
    merged_conf,
    geo,
    forcingformat,
    selected_converter,
    ref_height,
    first_base_time,
    timestep,
    system_file_paths=None,
):
    """Set the input parameter for a specific SURFEX forcing variable based on input.

    Args:
        sfx_var (str): _description_
        merged_conf (_type_): _description_
        geo (_type_): _description_
        forcingformat (_type_): _description_
        selected_converter (_type_): _description_
        ref_height (_type_): _description_
        first_base_time (_type_): _description_
        timestep (_type_): _description_
        system_file_paths(SystemFilePaths): System file paths

    Returns:
        _type_: _description_

    Raises:
        KeyError: KeyError

    """
    #########################################
    # 1. Gobal configuration from yaml file
    #########################################

    conf = copy.deepcopy(merged_conf)

    # Now we know the CLA settings for each surfex variable

    # Set defaults (merged global and user settings)
    # Theses must be sent to the converter to be used for each variable
    defs = {}
    if forcingformat in conf:
        defs = copy.deepcopy(conf[forcingformat])
        defs.update({"timestep": timestep})

    # Macros
    defs["system_file_paths"] = system_file_paths

    # All objects with converters, find converter dict entry
    conf_dict = {}
    if forcingformat != "constant":
        # Non-height dependent variables
        if ref_height is None:
            if "converter" in conf[sfx_var][forcingformat]:
                conf_dict = copy.deepcopy(conf[sfx_var][forcingformat]["converter"])
            else:
                raise KeyError("No converter defined for " + sfx_var)

        # Variables with height dependency
        elif ref_height in conf[sfx_var]:
            if forcingformat not in conf[sfx_var][ref_height]:
                msg = (
                    f"{conf[sfx_var]!s}: "
                    + f"Missing definitions for {sfx_var} and format: {forcingformat}"
                )
                raise KeyError(msg)
            if conf[sfx_var][ref_height][forcingformat] is None:
                raise KeyError(f"{conf[sfx_var]!s}: Missing definitions for {sfx_var}")
            if "converter" in conf[sfx_var][ref_height][forcingformat]:
                conf_dict = copy.deepcopy(
                    conf[sfx_var][ref_height][forcingformat]["converter"]
                )
            else:
                raise KeyError("No converter defined for " + sfx_var)
        else:
            raise KeyError('No ref height "' + ref_height + '" defined for ' + sfx_var)

    ##############################################################
    ##############################################################
    ##############################################################
    # Create the object to be returned
    if forcingformat == "constant":
        if ref_height is None:
            if "constant" in conf[sfx_var]:
                const_dict = copy.deepcopy(conf[sfx_var]["constant"])
            else:
                raise KeyError("No constant defined for " + sfx_var)
        elif ref_height in conf[sfx_var]:
            if "constant" in conf[sfx_var][ref_height]:
                const_dict = copy.deepcopy(conf[sfx_var][ref_height]["constant"])
            else:
                raise KeyError("No constant defined for " + sfx_var)
        else:
            raise KeyError('No ref height "' + ref_height + '" defined for ' + sfx_var)

        obj = ConstantValue(geo, sfx_var, const_dict)
    else:
        # Construct the converter
        converter = Converter(
            selected_converter, first_base_time, defs, conf_dict, forcingformat
        )

        # Construct the input object
        obj = ConvertedInput(geo, sfx_var, converter)
    return obj


def set_forcing_config(**kwargs):
    """Set the forcing config."""
    file_base = None
    domain_file = kwargs["domain"]
    with open(domain_file, mode="r", encoding="utf-8") as fhandler:
        geo_out = get_geo_object(json.load(fhandler))

    user_config = {}
    pattern = None
    timestep = 3600
    cache_interval = 3600
    zsoro = "default"
    zsoro_converter = "none"
    zval = "default"
    zval_converter = "none"
    uval = "default"
    uval_converter = "none"

    tair = "default"
    tair_converter = "none"
    qair = "default"
    qair_converter = "none"
    psurf = "default"
    psurf_converter = "none"
    dir_sw = "default"
    dir_sw_converter = "none"
    sca_sw = "default"
    sca_sw_converter = "none"
    lw_rad = "default"
    lw_rad_converter = "none"
    rain = "default"
    rain_converter = "none"
    snow = "default"
    snow_converter = "none"
    wind = "default"
    wind_converter = "none"
    wind_dir = "default"
    wind_dir_converter = "none"
    co2 = "default"
    co2_converter = "none"

    diskless_write = False
    analysis = False
    interpolation = None
    try:
        dtg_start = kwargs["dtg_start"]
        dtg_stop = kwargs["dtg_stop"]
        input_format = kwargs["input_format"]
        output_format = kwargs["output_format"]
        outfile = kwargs["output_filename"]
        zref = kwargs["zref"]
        uref = kwargs["uref"]
        config = kwargs["config"]
        if config is None:
            config_file = f"{PACKAGE_DIRECTORY}/cfg/config.yml"
            with open(config_file, mode="r", encoding="utf-8") as file_handler:
                config = yaml.safe_load(file_handler)

        if "interpolation" in kwargs:
            interpolation = kwargs["interpolation"]
        if "analysis" in kwargs:
            analysis = kwargs["analysis"]
        if "filebase" in kwargs:
            file_base = kwargs["filebase"]
        if "geo_out" in kwargs:
            geo_out = kwargs["geo_out"]
        if "user_config" in kwargs:
            user_config = kwargs["user_config"]
            if isinstance(user_config, str):
                with open(user_config, mode="r", encoding="utf-8") as file_handler:
                    user_config = yaml.safe_load(file_handler)
        if "pattern" in kwargs:
            pattern = kwargs["pattern"]
        if "timestep" in kwargs:
            timestep = kwargs["timestep"]
        if "cache_interval" in kwargs:
            cache_interval = kwargs["cache_interval"]
        if "zsoro" in kwargs:
            zsoro = kwargs["zsoro"]
        if "zsoro_converter" in kwargs:
            zsoro_converter = kwargs["zsoro_converter"]
        if "zval" in kwargs:
            zval = kwargs["zval"]
        if "zval_converter" in kwargs:
            zval_converter = kwargs["zval_converter"]
        if "uval_converter" in kwargs:
            uval_converter = kwargs["uval_converter"]
        if "uval" in kwargs:
            uval = kwargs["uval"]
        if "ta" in kwargs:
            tair = kwargs["ta"]
        if "ta_converter" in kwargs:
            tair_converter = kwargs["ta_converter"]
        if "qa" in kwargs:
            qair = kwargs["qa"]
        if "qa_converter" in kwargs:
            qair_converter = kwargs["qa_converter"]
        if "ps" in kwargs:
            psurf = kwargs["ps"]
        if "ps_converter" in kwargs:
            psurf_converter = kwargs["ps_converter"]
        if "dir_sw" in kwargs:
            dir_sw = kwargs["dir_sw"]
        if "dir_sw_converter" in kwargs:
            dir_sw_converter = kwargs["dir_sw_converter"]
        if "sca_sw" in kwargs:
            sca_sw = kwargs["sca_sw"]
        if "sca_sw_converter" in kwargs:
            sca_sw_converter = kwargs["sca_sw_converter"]
        if "lw" in kwargs:
            lw_rad = kwargs["lw"]
        if "lw_converter" in kwargs:
            lw_rad_converter = kwargs["lw_converter"]
        if "rain" in kwargs:
            rain = kwargs["rain"]
        if "rain_converter" in kwargs:
            rain_converter = kwargs["rain_converter"]
        if "snow" in kwargs:
            snow = kwargs["snow"]
        if "snow_converter" in kwargs:
            snow_converter = kwargs["snow_converter"]
        if "wind" in kwargs:
            wind = kwargs["wind"]
        if "wind_converter" in kwargs:
            wind_converter = kwargs["wind_converter"]
        if "wind_dir" in kwargs:
            wind_dir = kwargs["wind_dir"]
        if "wind_dir_converter" in kwargs:
            wind_dir_converter = kwargs["wind_dir_converter"]
        if "co2" in kwargs:
            co2 = kwargs["co2"]
        if "co2_converter" in kwargs:
            co2_converter = kwargs["co2_converter"]
        if "diskless_write" in kwargs:
            diskless_write = kwargs["diskless_write"]

    except ValueError:
        raise RuntimeError("Needed input is missing") from ValueError

    # Time information
    if (int(dtg_start) or int(dtg_stop)) < 1000010100:
        raise RuntimeError(
            "Invalid start and stop times! " + str(dtg_start) + " " + str(dtg_stop)
        )

    start = as_datetime(str.strip(str(dtg_start)))
    stop = as_datetime(str.strip(str(dtg_stop)))
    if file_base is None:
        first_base_time = start
    else:
        first_base_time = as_datetime(str.strip(str(file_base)))

    logging.info("config=%s", config)
    logging.info("user_config=%s", user_config)
    # Merge all settings with user all settings
    merged_conf = deep_update(config, user_config)
    logging.info("merged_conf=%s", merged_conf)

    logging.info("fileformat=%s, pattern=%s", input_format, pattern)

    # Replace global settings from
    fileformat = input_format
    if fileformat not in merged_conf:
        merged_conf.update({fileformat: {}})
    if pattern is not None:
        merged_conf[fileformat]["filepattern"] = pattern

    # Interpolation
    merged_conf[fileformat]["interpolation"] = interpolation

    # Prefer forecast or analysis
    if analysis:
        merged_conf[fileformat]["prefer_forecast"] = False
        merged_conf[fileformat]["fcint"] = 3600.0
        merged_conf[fileformat]["offset"] = 0

    geo_input = None
    if "geo_input" in kwargs:
        geo_input = kwargs["geo_input"]
        merged_conf[fileformat]["geo_input"] = geo_input
    if geo_input is None and "geo_input_file" in kwargs:
        geo_input_file = kwargs["geo_input_file"]
        if geo_input_file is not None:
            if os.path.exists(geo_input_file):
                with open(geo_input_file, "r", encoding="utf-8") as fhandler:
                    geo_input = get_geo_object(json.load(fhandler))
                merged_conf[fileformat]["geo_input_file"] = geo_input_file
            else:
                logging.warning("Input geometry %s does not exist", geo_input_file)

    try:
        system_file_paths = kwargs["system_file_paths"]
    except KeyError:
        system_file_paths = None

    # Set attributes
    atts = ["ZS", "ZREF", "UREF"]
    att_objs = []
    for att_var in atts:
        logging.info("Set variable %s", att_var)
        # Override with command line options for a given variable
        ref_height = None
        cformat = fileformat
        if att_var == "ZS":
            if zsoro != "default":
                cformat = zsoro
            selected_converter = zsoro_converter
        elif att_var == "ZREF":
            if zval != "default":
                cformat = zval
            selected_converter = zval_converter
            ref_height = zref
            if ref_height == "screen":
                cformat = "constant"
        elif att_var == "UREF":
            if uval != "default":
                cformat = uval
            selected_converter = uval_converter
            ref_height = uref
            if ref_height == "screen":
                cformat = "constant"
        else:
            raise NotImplementedError

        att_objs.append(
            set_input_object(
                att_var,
                merged_conf,
                geo_out,
                cformat,
                selected_converter,
                ref_height,
                first_base_time,
                timestep,
                system_file_paths=system_file_paths,
            )
        )

    # Set forcing variables (time dependent)
    variables = [
        "TA",
        "QA",
        "PS",
        "DIR_SW",
        "SCA_SW",
        "LW",
        "RAIN",
        "SNOW",
        "WIND",
        "WIND_DIR",
        "CO2",
    ]
    var_objs = []
    # Search in config file for parameters to override
    for sfx_var in variables:
        logging.info("Set variable %s", sfx_var)
        ref_height = None
        cformat = fileformat
        if sfx_var == "TA":
            if tair != "default":
                cformat = tair
            selected_converter = tair_converter
            ref_height = zref
        elif sfx_var == "QA":
            if qair != "default":
                cformat = qair
            selected_converter = qair_converter
            ref_height = zref
        elif sfx_var == "PS":
            if psurf != "default":
                cformat = psurf
            selected_converter = psurf_converter
        elif sfx_var == "DIR_SW":
            if dir_sw != "default":
                cformat = dir_sw
            selected_converter = dir_sw_converter
        elif sfx_var == "SCA_SW":
            if sca_sw != "default":
                cformat = sca_sw
            selected_converter = sca_sw_converter
        elif sfx_var == "LW":
            if lw_rad != "default":
                cformat = lw_rad
            selected_converter = lw_rad_converter
        elif sfx_var == "RAIN":
            if rain != "default":
                cformat = rain
            selected_converter = rain_converter
        elif sfx_var == "SNOW":
            if snow != "default":
                cformat = snow
            selected_converter = snow_converter
        elif sfx_var == "WIND":
            if wind != "default":
                cformat = wind
            selected_converter = wind_converter
            ref_height = uref
        elif sfx_var == "WIND_DIR":
            if wind_dir != "default":
                cformat = wind_dir
            selected_converter = wind_dir_converter
            ref_height = uref
        elif sfx_var == "CO2":
            if co2 != "default":
                cformat = co2
            selected_converter = co2_converter
        else:
            raise NotImplementedError
        var_objs.append(
            set_input_object(
                sfx_var,
                merged_conf,
                geo_out,
                cformat,
                selected_converter,
                ref_height,
                first_base_time,
                timestep,
                system_file_paths=system_file_paths,
            )
        )

    # Save options
    options = {}
    options["output_format"] = output_format
    options["output_file"] = outfile
    options["diskless_write"] = diskless_write
    options["start"] = start
    options["stop"] = stop
    options["timestep"] = timestep
    options["geo_out"] = geo_out
    options["single"] = False
    options["force"] = kwargs["force"]
    if "single" in kwargs:
        options["single"] = kwargs["single"]
    options["cache_interval"] = cache_interval

    return options, var_objs, att_objs


def modify_forcing(**kwargs):
    """Modify forcing."""
    infile = kwargs["input_file"]
    outfile = kwargs["output_file"]
    time_step = kwargs["time_step"]
    variables = kwargs["variables"]

    ifile = netCDF4.Dataset(infile, "r")
    ofile = netCDF4.Dataset(outfile, "r+")

    for var in variables:
        logging.info("Modify variable %s", var)
        logging.info(
            "input %s %s %s",
            ifile[var][time_step, :],
            ifile[var][time_step, :].shape,
            time_step,
        )
        logging.info("output %s %s", ofile[var][0, :], ofile[var][0, :].shape)
        ofile[var][0, :] = ifile[var][time_step, :]
        ofile.sync()

    ifile.close()
    ofile.close()



"""FA support."""

import logging

try:
    from epygram.formats import resource
except ImportError:
    resource = None


from .geo import ConfProj
from .interpolation import Interpolation


class Fa(object):
    """Fichier Arpege."""

    def __init__(self, fname):
        """Construct a FA object.

        Args:
            fname (str): filename
        """
        self.fname = fname
        self.projection = None
        self.lons = None
        self.lats = None
        self.nearest = None
        self.linear = None

    def field(self, varname, validtime):
        """Read a field.

        Args:
            varname (_type_): _description_
            validtime (_type_): _description_

        Raises:
            ModuleNotFoundError: You need epygram to read FA files
            NotImplementedError: Geometry not implemented

        Returns:
            tuple: np.field, surfex.Geometry

        """
        if resource is None:
            raise ModuleNotFoundError("You need epygram to read FA files")

        fa_file = resource(self.fname, openmode="r")
        field = fa_file.readfield(varname)
        # TODO this might not work with forcing...
        zone = "CI"
        crnrs = field.geometry.gimme_corners_ij(subzone=zone)

        range_x = slice(crnrs["ll"][0], crnrs["lr"][0] + 1)
        range_y = slice(crnrs["lr"][1], crnrs["ur"][1] + 1)
        # TODO: check time
        logging.info(
            "Not checking validtime for FA variable at the moment: %s", str(validtime)
        )

        if field.geometry.name in ("lambert", "polar_stereographic"):
            n_y = field.geometry.dimensions["Y_CIzone"]
            n_x = field.geometry.dimensions["X_CIzone"]
            lon0 = field.geometry.projection["reference_lon"].get("degrees")
            lat0 = field.geometry.projection["reference_lat"].get("degrees")
            c0, c1 = field.geometry.getcenter()
            lonc = c0.get("degrees")
            latc = c1.get("degrees")
            d_x = field.geometry.grid["X_resolution"]
            d_y = field.geometry.grid["Y_resolution"]
            ilone = field.geometry.dimensions["X"] - n_x
            ilate = field.geometry.dimensions["Y"] - n_y

            domain = {
                "nam_conf_proj": {"xlon0": lon0, "xlat0": lat0},
                "nam_conf_proj_grid": {
                    "xloncen": lonc,
                    "xlatcen": latc,
                    "nimax": n_x,
                    "njmax": n_y,
                    "xdx": d_x,
                    "xdy": d_y,
                    "ilone": ilone,
                    "ilate": ilate,
                },
            }
            geo_out = ConfProj(domain)
            if field.spectral:
                field.sp2gp()
            data = field.data[range_y, range_x].T
        else:
            raise NotImplementedError(field.geometry.name + " not implemented yet!")
        return data, geo_out

    def points(self, varname, geo, validtime=None, interpolation="nearest"):
        """Read a 2-D field and interpolates it to requested positions.

        Args:
            varname (str): Variable name
            geo (surfex.geo.Geo): Geometry
            validtime (as_datetime): Validtime
            interpolation (str, optional): Interpoaltion method. Defaults to "nearest".

        Returns:
            tuple: field, interpolator

        """
        field, geo_in = self.field(varname, validtime)
        interpolator = Interpolation(interpolation, geo_in, geo)

        field = interpolator.interpolate(field)
        return field, interpolator



"""Variable."""

import copy
import logging

import numpy as np

from .datetime_utils import as_timedelta
from .fa import Fa
from .file import SurfexFileVariable, get_surfex_io_object
from .geo import get_geo_object_from_json_file
from .grib import Grib, Grib1Variable, Grib2Variable
from .input_methods import get_datasources
from .interpolation import Interpolation
from .netcdf import Netcdf, NetCDFReadVariable
from .platform_deps import SystemFilePathsFromFile
from .util import parse_filepattern


class Variable(object):
    """New combined variable."""

    def __init__(self, var_type, var_dict, initial_basetime, prefer_forecast=True):
        """Construct variable.

        Args:
            var_type (str): Variable type.
            var_dict (dict): Variable definitions
            initial_basetime (datetime): Initial basetime
            prefer_forecast (bool, optional): Prefer forecasts instead of analysis.
                                              Defaults to True.

        Raises:
            RuntimeError: No filepattern provided
            RuntimeError: variable must have attribute
            RuntimeError: You can not have larger offset than the frequency of forecasts

        """
        self.var_type = var_type
        self.var_dict = copy.deepcopy(var_dict)
        try:
            self.interval = var_dict["timestep"]
        except KeyError:
            self.interval = 3600

        self.filepattern = self.substitute_macros(var_dict["filepattern"])
        self.initial_basetime = initial_basetime
        try:
            self.fcint = int(self.var_dict["fcint"])
        except KeyError:
            self.fcint = 10800
        try:
            self.offset = int(self.var_dict["offset"])
        except KeyError:
            self.offset = 0
        try:
            self.prefer_forecast = self.var_dict["prefer_forecast"]
        except KeyError:
            self.prefer_forecast = prefer_forecast

        try:
            self.one_forecast = self.var_dict["one_forecast"]
        except KeyError:
            self.one_forecast = False

        if self.offset > self.fcint:
            raise RuntimeError(
                "You can not have larger offset than the frequency of forecasts "
                + str(self.offset)
                + " > "
                + str(self.fcint)
            )
        accumulated, instant, file_var = self.set_var(validtime=self.initial_basetime)
        self.file_var = file_var
        self.instant = instant
        self.accumulated = accumulated
        self.alpha = None
        logging.debug("Constructed variable for %s", str(self.var_dict))

    def substitute_macros(self, setting, micro="@"):
        """Substitute macros.

        Args:
            setting (str): Setting
            micro (str): Micro character

        """
        try:
            spaths = self.var_dict["system_file_paths"]
        except KeyError:
            logging.info("No system file paths used")
            spaths = None
        if spaths is None:
            return setting

        spaths = SystemFilePathsFromFile(spaths)
        if isinstance(setting, str):
            for mkey, mvalue in spaths.system_file_paths.items():
                logging.debug("mkey: %s mvalue=%s", mkey, mvalue)
                if isinstance(mkey, str) and isinstance(mvalue, str):
                    logging.debug("mkey: %s mvalue=%s", mkey, mvalue)
                    setting = setting.replace(f"{micro}{mkey}{micro}", mvalue)
        return setting

    def get_filename(self, validtime, previoustime=None):
        """Get the filename.

        Args:
            validtime (datetime.datetime): Valid time.
            previoustime (datetime.datetime, optional): Previous valid time.
                                                        Defaults to None.

        Returns:
            str: Parsed filename

        """
        logging.debug("Set basetime for filename")
        basetime = None
        if validtime is not None:
            basetime = self.get_basetime(validtime, previoustime=previoustime)
            if previoustime is not None:
                validtime = previoustime
        return parse_filepattern(self.filepattern, basetime, validtime)

    def get_filehandler(self, validtime, cache=None, previoustime=None):
        """Get the file handler.

        Args:
            validtime (datetime.datetime): Valid time.
            cache (surfex.Cache, optional): Cache. Defaults to None.
            previoustime (datetime.datetime, optional): Previous valid time.
                                                        Defaults to None.

        Raises:
            NotImplementedError: Variable type not implemented

        Returns:
            tuple: Filehandler and file name

        """
        logging.info("validtime, %s previoustime %s", validtime, previoustime)
        filename = self.get_filename(validtime, previoustime=previoustime)
        if cache is not None and cache.file_open(filename):
            file_handler = cache.get_file_handler(filename)
        else:
            if self.var_type == "netcdf":
                file_handler = Netcdf(filename)
            elif self.var_type in ("grib1", "grib2"):
                file_handler = Grib(filename)
            elif self.var_type == "fa":
                file_handler = Fa(filename)
            elif self.var_type == "surfex":
                try:
                    fileformat = self.var_dict["fileformat"]
                except KeyError:
                    fileformat = None
                try:
                    filetype = self.var_dict["filetype"]
                except KeyError:
                    filetype = None
                geo_in = None
                if "geo_input" in self.var_dict:
                    geo_in = self.var_dict["geo_input"]
                elif "geo_input_file" in self.var_dict:
                    geo_in_file = self.var_dict["geo_input_file"]
                    geo_in = get_geo_object_from_json_file(geo_in_file)

                file_handler = get_surfex_io_object(
                    filename, fileformat=fileformat, filetype=filetype, geo=geo_in
                )
            elif self.var_type == "obs":
                var_dict = self.var_dict
                var_dict = {"set": var_dict}
                file_handler = get_datasources(validtime, var_dict)[0]
            else:
                raise NotImplementedError

            if cache is not None:
                cache.set_file_handler(filename, file_handler)

        logging.debug("var_type: %s", self.var_type)
        logging.debug("filename: %s", filename)
        return file_handler, filename

    def read_var_field(self, validtime, cache=None):
        """Read points for a variable.

        Args:
            validtime (datetime.datetime): Valid time
            cache (surfex.Cache, optional): Cache. Defaults to None.

        Returns:
            numpy.darray: A numpy array with point values for variable

        Raises:
            NotImplementedError: "Field not defined for point data"

        """
        logging.debug("set basetime from %s", validtime)

        filehandler, filename = self.get_filehandler(validtime, cache=cache)

        id_str = None
        if cache is not None:
            id_str = cache.generate_id(self.var_type, self.file_var, filename, validtime)

        geo_read = None
        if cache is not None and cache.is_saved(id_str):
            field = cache.saved_fields[id_str]
            logging.info("Using cached value for %s", id_str)
        else:
            if self.var_type == "obs":
                raise NotImplementedError("Field not defined for point data")
            field, geo_read = filehandler.field(self.file_var, validtime=validtime)

            if field is not None:
                logging.debug("field.shape %s", field.shape)

            if cache is not None:
                logging.debug("ID_STRING %s", id_str)
                cache.save_field(id_str, field)
        return field, geo_read

    def read_var_points(self, var, geo, validtime, previoustime=None, cache=None):
        """Read points for a variable.

        Args:
            var (object): Variable object
            geo (surfex.Geo): Surfex geometry
            validtime (datetime.datetime): Valid time
            previoustime (datetime.datetime, optional): Previous valid time for
                                                        accumulated variables.
                                                        Defaults to None.
            cache (surfex.Cache, optional): Cache. Defaults to None.

        Returns:
            numpy.darray: A numpy array with point values for variable

        """
        interpolation = "bilinear"
        if self.var_type != "obs" and "interpolator" in self.var_dict:
            interpolation = self.var_dict["interpolator"]

        logging.debug("set basetime from %s", validtime)
        filehandler, filename = self.get_filehandler(
            validtime, cache=cache, previoustime=previoustime
        )
        if previoustime is not None:
            validtime = previoustime
            logging.debug("new validtime to read previous time %s", validtime)

        if filehandler is None:
            logging.warning("No file handler exist for this time step")
            field = np.array([len(geo.lons)])
            field = field.fill(np.nan)
        else:
            id_str = None
            if cache is not None:
                id_str = cache.generate_id(self.var_type, var, filename, validtime)

            if cache is not None and cache.is_saved(id_str):
                field = cache.saved_fields[id_str]
                logging.info("Using cached value for %s", id_str)
            else:
                if self.var_type == "obs":
                    if geo is None:
                        raise RuntimeError("No geo is provided!")
                    __, field, __ = filehandler.points(geo, validtime=validtime)
                else:
                    field, interpolator = filehandler.points(
                        var, geo, interpolation=interpolation, validtime=validtime
                    )
                    # Set alpha if requested
                    self.rotate_geographic_wind(interpolator)
                if field is not None:
                    logging.debug("field.shape %s", field.shape)

                if cache is not None:
                    logging.debug("ID_STRING %s", id_str)
                    cache.save_field(id_str, field)
        return field

    def set_var(self, validtime=None):
        """Set the variable.

        Args:
            validtime (datetime.datetime, optional): Valid time. Defaults to None.

        Raises:
            NotImplementedError: Variable not implemented
            RuntimeError: You must provide name
            RuntimeError: grib1 needs type or levelType

        Returns:
            tuple: accumulated, instant, var

        """
        accumulated = False
        if self.var_type == "netcdf":
            try:
                name = self.var_dict["name"]
            except KeyError:
                raise RuntimeError("You must provide name") from KeyError
            try:
                accumulated = self.var_dict["accumulated"]
            except KeyError:
                accumulated = False
            try:
                level = self.var_dict["level"]
                if not isinstance(level, list):
                    level = [level]
            except KeyError:
                level = None
            try:
                units = str([self.var_dict["units"]][0])
            except KeyError:
                units = None
            try:
                member = self.var_dict["member"]
                if member is not None and not isinstance(member, list):
                    member = [member]
            except KeyError:
                member = None

            var = NetCDFReadVariable(name, level=level, units=units, member=member)
        elif self.var_type == "grib1":
            try:
                par = self.var_dict["parameter"]
            except KeyError:
                raise RuntimeError("You must provide parameter") from KeyError
            try:
                typ = self.var_dict["type"]
            except KeyError:
                try:
                    typ = self.var_dict["levelType"]
                except KeyError:
                    raise RuntimeError("grib1 needs type or levelType") from KeyError

            try:
                level = self.var_dict["level"]
            except KeyError:
                raise RuntimeError("grib1 needs level") from KeyError
            try:
                tri = self.var_dict["tri"]
            except KeyError:
                tri = 0
            if tri == 4:
                accumulated = True
            var = Grib1Variable(par, typ, level, tri=tri)
        elif self.var_type == "grib2":
            discipline = self.var_dict["discipline"]
            p_c = self.var_dict["parameterCategory"]
            p_n = self.var_dict["parameterNumber"]
            l_t = self.var_dict["levelType"]
            lev = self.var_dict["level"]
            try:
                tsp = self.var_dict["typeOfStatisticalProcessing"]
            except KeyError:
                tsp = -1
            if tsp == 1:
                accumulated = True
            var = Grib2Variable(discipline, p_c, p_n, l_t, lev, tsp=tsp)
        elif self.var_type == "surfex":
            varname = self.var_dict["varname"]
            layers = None
            if "layers" in self.var_dict:
                layers = self.var_dict["layers"]
            patches = None
            if "patches" in self.var_dict:
                patches = self.var_dict["patches"]
            if "accumulated" in self.var_dict:
                accumulated = self.var_dict["accumulated"]
            datatype = "float"
            if "datatype" in self.var_dict:
                datatype = self.var_dict["datatype"]
            tiletype = "FULL"
            if "tiletype" in self.var_dict:
                tiletype = self.var_dict["tiletype"]

            basetime = None
            if validtime is not None:
                basetime = self.get_basetime(validtime)
            var = SurfexFileVariable(
                varname,
                validtime=validtime,
                patches=patches,
                layers=layers,
                basetime=basetime,
                interval=self.interval,
                datatype=datatype,
                tiletype=tiletype,
            )
        elif self.var_type == "fa":
            var = self.var_dict["name"]
            if "accumulated" in self.var_dict:
                accumulated = self.var_dict["accumulated"]

        elif self.var_type == "obs":
            var = self.var_dict.get("varname", None)
        else:
            raise NotImplementedError

        instant = 0
        if accumulated:
            instant = self.interval
        if "instant" in self.var_dict:
            instant = self.var_dict["instant"]
        if "prefer_forecast" in self.var_dict:
            self.prefer_forecast = self.var_dict["prefer_forecast"]

        return accumulated, instant, var

    def read_variable(self, geo, validtime, cache=None):
        """Read the variable.

        Args:
            geo (surfex.Geometry): Geometry to read
            validtime (datetime.datetime): Valid time.
            cache (surfex.cache): Cache. Defaults to None

        Raises:
            RuntimeError: Negative accumulated value found

        Returns:
            np.darray: Field read and interpolated.

        """
        # Set variable info
        previous_field = None
        first_time = False
        if self.accumulated:
            # Re-read field
            previoustime = validtime - as_timedelta(seconds=self.interval)
            # Don't read if previous time is older than the very first basetime
            if previoustime >= self.initial_basetime:
                logging.debug("Re-read %s", previoustime)
                previous_field = self.read_var_points(
                    self.file_var,
                    geo,
                    validtime=validtime,
                    previoustime=previoustime,
                    cache=cache,
                )
            else:
                first_time = True
                previous_field = np.zeros([geo.npoints])
        elif self.instant > 0:
            previous_field = np.zeros([geo.npoints])

        deaccumulate = False
        if self.accumulated or self.instant > 0:
            deaccumulate = True

        # Always set accumulated values to zero the first time
        if first_time:
            deaccumulate = False
            field = np.zeros([geo.npoints])
        else:
            field = self.read_var_points(
                self.file_var, geo, validtime=validtime, cache=cache
            )

        # Deaccumulate if either two files are read or if instant is > 0.
        if deaccumulate:
            field = self.deaccumulate(field, previous_field, self.instant)
            if any(field[field < 0.0]):
                raise RuntimeError(
                    "Negative accumulated value found for " + self.file_var
                )
        return field

    def print_variable_info(self):
        """Print variable."""
        logging.debug(":%s:", str(self.var_dict))

    def deaccumulate(self, field, previous_field, instant):
        """Deaccumulate variable.

        Args:
            field (_type_): _description_
            previous_field (_type_): _description_
            instant (_type_): _description_

        Raises:
            ValueError: _description_

        Returns:
            _type_: _description_

        """
        logging.debug("Deaccumulate field with: %s", instant)
        if field is None:
            logging.warning("Field is not read properly")
            return None

        field = np.subtract(field, previous_field)
        if any(field[field < 0.0]):
            neg = []
            for i in range(field.shape[0]):
                if field[i] < 0.0:
                    neg.append(field[i])
            neg = np.asarray(neg)
            logging.warning(
                "Deaccumulated field has %s negative values. lowest: %s mean: %s",
                str(neg.shape[0]),
                str(np.nanmin(neg)),
                str(np.nanmean(neg)),
            )
        field[field < 0.0] = 0
        if any(field[field < 0.0]):
            raise ValueError("Should not be negative values")
        if float(instant) != 0.0:
            field = np.divide(field, float(instant))

        return field

    def get_basetime(self, validtime, previoustime=None, allow_different_basetime=False):
        """Get the basetime of the file.

        Args:
            validtime (datetime.datetime): Valid time
            previoustime (datetime.datetime, optional): Previous valid time.
                                                        Defaults to None.
            allow_different_basetime (bool, optional): Allow different base times.
                                                       Defaults to False.

        Raises:
            RuntimeError: "Negative offset does not make sense here"
            NotImplementedError: _description_

        Returns:
            datetime.datetime: Base time.

        """
        if self.offset < 0:
            raise RuntimeError("Negative offset does not make sense here")

        if self.one_forecast:
            logging.debug(
                "one_forecast is True. Valitime %s Keep initial basetime %s",
                validtime,
                self.initial_basetime,
            )
            return self.initial_basetime

        # Take offset into account
        first = False
        offset = self.offset

        basetime = validtime - as_timedelta(seconds=offset)
        if basetime <= self.initial_basetime:
            first = True
            basetime = self.initial_basetime

        logging.debug("             validtime: %s", validtime)
        logging.debug("          previoustime: %s", previoustime)
        logging.debug("                 first: %s", first)
        logging.debug("                 fcint: %s", self.fcint)
        logging.debug("                offset: %s", self.offset)
        logging.debug("      initial_basetime: %s", self.initial_basetime)
        logging.debug("  Basetime with offset: %s", basetime)

        # Modify based on fcint
        seconds_since_midnight = int(
            (
                basetime - basetime.replace(hour=0, minute=0, second=0, microsecond=0)
            ).total_seconds()
        )
        if seconds_since_midnight == 86400:
            seconds_since_midnight = 0
        basetime_inc = int(
            seconds_since_midnight / int(as_timedelta(seconds=self.fcint).total_seconds())
        )

        prefer_forecast = as_timedelta(seconds=0)
        if seconds_since_midnight == basetime_inc * int(
            as_timedelta(seconds=self.fcint).seconds
        ):
            if first:
                logging.debug("First basetime")
            elif self.prefer_forecast:
                logging.debug("Prefer forecasts instead of analyis")
                prefer_forecast = as_timedelta(seconds=self.fcint)
            elif not self.prefer_forecast:
                logging.debug("Prefer analysis instead of forecast")

        fcint = as_timedelta(seconds=self.fcint)
        basetime = (
            basetime.replace(hour=0, minute=0, second=0, microsecond=0)
            + (basetime_inc * fcint)
            - prefer_forecast
        )

        if previoustime is not None and allow_different_basetime:
            raise NotImplementedError

        logging.debug("seconds_since_midnight: %s", seconds_since_midnight)
        logging.debug(
            "         cycle seconds: %s",
            basetime_inc * int(as_timedelta(seconds=self.fcint).seconds),
        )
        logging.debug("       prefer_forecast: %s", self.prefer_forecast)
        logging.debug("          one_forecast: %s", self.one_forecast)
        logging.debug("   prefer_forecast_inc: %s", prefer_forecast)
        logging.debug("          basetime_inc: %s", basetime_inc)
        logging.debug("           Basetime is: %s", basetime)
        return basetime

    def rotate_geographic_wind(self, interpolator):
        """Rotate wind.

        Args:
            interpolator (_type_): _description_

        """
        rotate_wind = False
        if "rotate_to_geographic" in self.var_dict:
            rotate_wind = self.var_dict["rotate_to_geographic"]
        if rotate_wind:
            alpha_grid = interpolator.alpha_grid_rot()
            alpha_interpolation = Interpolation(
                "nearest", interpolator.geo_in, interpolator.geo_out
            )
            self.alpha = alpha_interpolation.interpolate(alpha_grid)



"""Namelist."""

import logging
import os
import re

import f90nml
import yaml

from .binary_input import InputDataFromNamelist


class NamelistGenerator(object):
    """Namelist class."""

    def __init__(self, program, nml, macros=None, micro="@"):
        """Construct a base namelists class.

        Args:
            program (str): Which surfex binary you want to run
                           ["pgd", "prep", "offline", "soda"]
            nml (f90nml.Namelist): A parsed fortran namelist
            assemble(dict): Assembly order. Defines the configuration
            macros(dict, optional): Macros
            micro(str, optional): Micro character

        """
        self.program = program
        self.nml = nml
        if macros is None:
            macros = {}
        self.macros = macros
        self.micro = micro

    def get_namelist(self):
        """Get namelist."""
        for bkey, block in self.nml.items():
            items = {}
            for key, val in block.items():
                rval = val
                items.update({key: rval})
                for mkey, mval in self.macros.items():
                    if isinstance(rval, str):
                        rval = rval.replace(f"{self.micro}{mkey}{self.micro}", f"{mval}")
                        if rval == f"{mval}" and isinstance(mval, (float, int)):
                            rval = mval
                    items.update({key: rval})
            self.nml[bkey].update(items)
        return self.nml

    def input_data_from_namelist(
        self, input_data, platform, basetime=None, validtime=None, check_existence=True
    ):
        """Construct a base namelists class to be implemented by namelist implementations.

        Args:
            input_data (dict): Input data definitions
            platform (SystemFilePaths): Platform specific settings
            basetime (as_datetime, optional): Base time
            validtime (as_datetime, optional): Valid time
            check_existence (bool, optional): Check existence of input data.
                                              Defaults to True
        Returns:
            data_obj (InputDataFromNamelist): Input data from namelist

        """
        logging.info("Set input data from namelist for program: %s", self.program)

        def process_macros(data):
            for k, v in data.copy().items():
                if isinstance(v, dict):  # For DICT
                    data[k] = process_macros(v)
                elif isinstance(v, list):  # For LIST
                    data[k] = [process_macros(i) for i in v]
                else:
                    vv = v
                    for macro, mval in self.macros.items():
                        if isinstance(vv, str) and isinstance(mval, str):
                            vv = vv.replace(f"{self.micro}{macro}{self.micro}", mval)
                    data[k] = vv
            return data

        # Substitute macros in binary input data
        if self.macros is not None:
            input_data = process_macros(input_data)

        data_obj = InputDataFromNamelist(
            self.nml,
            input_data,
            self.program,
            platform,
            basetime=basetime,
            validtime=validtime,
        )

        if check_existence:
            for val in data_obj.data.values():
                if not os.path.exists(val):
                    raise FileNotFoundError(val)
        return data_obj

    def write(
        self, output_file, uppercase=True, true_repr=".TRUE.", false_repr=".FALSE."
    ):
        """Generate the namelists for 'target'.

        Args:
            output_file (str): where to write the result
            (OPTIONS.nam, fort.4 or EXSEG1.nam typically)
            uppercase (bool, optional): Upper case namelist. Default to True
            true_repr (str, optional): String representation of fortran boolean true.
                                       Defaults to ".TRUE."
            false_repr (str, optional): String representation of fortran boolean false.
                                        Defaults to ".FALSE."

        """
        self.nml.uppercase = uppercase
        self.nml.true_repr = true_repr
        self.nml.false_repr = false_repr
        self.nml.write(output_file, force=True)
        logging.debug("Wrote: %s", output_file)

    @staticmethod
    def lower_case_namelist_dict(dict_in):
        """Lower case namelist.

        Args:
            dict_in (dict): Namelist dictionary to lower case

        Returns:
            dict: Namelist in lower case

        """
        new_dict = {}
        for key in dict_in:
            lower_case_dict = {}
            for key2 in dict_in[key]:
                lower_case_dict.update({key2.lower(): dict_in[key][key2]})
            new_dict.update({key.lower(): lower_case_dict})
        return new_dict


class NamelistGeneratorFromNamelistFile(NamelistGenerator):
    """Namelist class."""

    def __init__(self, program, nml_file, macros=None, micro="@"):
        """Construct a base namelists class from a file.

        Args:
            program (str): Which surfex binary you want to run
                           ["pgd", "prep", "offline", "soda"]
            nml_file (str): Namelist file
            macros(dict, optional): Macros
            micro(str, optional): Micro character

        """
        parser = f90nml.Parser()
        nml = parser.read(nml_file)
        NamelistGenerator.__init__(self, program, nml, macros=macros, micro=micro)


class NamelistGeneratorAssemble(NamelistGenerator):
    """Namelist class."""

    def __init__(self, program, definitions, assemble, macros=None, micro="@"):
        """Construct a base namelists class.

        Args:
            program (str): Which surfex binary you want to run
                           ["pgd", "prep", "offline", "soda"]
            definitions (dict): Namelist definitions
            assemble(dict): Assembly order. Defines the configuration
            macros(dict, optional): Macros
            micro(str, optional): Micro character

        """
        self.nldict = definitions
        if macros is None:
            macros = {}
        self.macros = macros

        self.assemble = assemble
        logging.info("Namelist blocks for program %s: %s", program, self.assemble)
        nlres = self.assemble_namelist(program)
        nml = f90nml.Namelist(nlres)
        NamelistGenerator.__init__(self, program, nml, macros=macros, micro=micro)

    def assemble_namelist(self, program):
        """Generate the namelists for 'target'.

        Args:
            program (str): Which surfex binary you want to run
                           ["pgd", "prep", "offline", "soda"]

        Raises:
            KeyError: Key not found

        Returns:
            nlres (dict): Assembled namelist

        """
        # Read namelist file with all the categories

        # Check target is valid
        cndict = self.assemble[program]
        nldict = self.nldict

        # Start with empty result dictionary
        nlres = {}

        # Assemble the target namelists based on the given category order
        for item in self.flatten_list(cndict):
            catg = item
            # variable substitution removed at this level (may be resurrected)
            # assemble namelists for this category
            if catg in nldict:
                for nl in nldict[catg]:
                    logging.debug("nl=%s catg=%s", nl, catg)
                    if nl not in nlres:
                        # create the result namelist dict
                        nlres[nl] = {}
                    if catg == "rm{" + nl + "}":
                        # clear/remove the given namelist (but not used for now)
                        nlres[nl].clear()
                    else:
                        for key in nldict[catg][nl]:
                            val = nldict[catg][nl][key]
                            finval = val
                            # Replace ${var-def} with value from config, possibly
                            # macro-expanded. For now assumes only one subst. per line,
                            # could be generalized if needed
                            if str(finval).find("$") >= 0:
                                m = re.search(
                                    r"^([^\$]*)\$\{([\w\#]+)\-?([^}]*)\}(.*)", str(val)
                                )
                                if m:
                                    pre = m.group(1)
                                    nam = m.group(2)
                                    defval = m.group(3)
                                    post = m.group(4)
                                    logging.debug("macros=%s", self.macros)
                                    logging.debug("look for nam=%s", nam)
                                    try:
                                        repval = self.macros[nam]
                                    except KeyError:
                                        repval = None
                                    if repval is None:
                                        if defval != "":
                                            logging.debug(
                                                "Using default value %s for '%s'",
                                                defval,
                                                nam,
                                            )
                                            repval = self.find_num(defval)
                                        else:
                                            logging.debug("No value found for: '%s'", nam)
                                    else:
                                        logging.debug(
                                            "Replaced %s with: %s", nam, str(repval)
                                        )
                                    if isinstance(repval, str):
                                        finval = str(pre) + str(repval) + str(post)
                                    else:
                                        finval = repval
                                else:
                                    raise KeyError(val)
                            nlres[nl][key] = finval
                            logging.debug("nl=%s key=%s, finval=%s", nl, key, finval)
            else:
                logging.info("Category %s not found in definitions", catg)
        return nlres

    @staticmethod
    def flatten_list(li):
        """Recursively flatten a list of lists (of lists)."""
        if li == []:
            return li
        if isinstance(li[0], list):
            return NamelistGeneratorAssemble.flatten_list(
                li[0]
            ) + NamelistGeneratorAssemble.flatten_list(li[1:])
        return li[:1] + NamelistGeneratorAssemble.flatten_list(li[1:])

    @staticmethod
    def find_num(s):
        """Purpose: un-quote numbers."""
        try:
            i = int(s)
            return i
        except ValueError:
            pass
        try:
            f = float(s)
            return f
        except ValueError:
            return s


class NamelistGeneratorAssembleFromFiles(NamelistGeneratorAssemble):
    """Namelist class."""

    def __init__(self, program, definitions, assemble, macros=None, micro="@"):
        """Construct a base namelists class.

        Args:
            program (str): Which surfex binary you want to run
                           ["pgd", "prep", "offline", "soda"]
            definitions (str): Namelist definitions
            assemble(str)): Assembly order. Defines the configuration
            macros(dict, optional): Macros
            micro(str, optional): Micro character

        """
        with open(definitions, mode="r", encoding="utf8") as file_handler:
            definitions = yaml.safe_load(file_handler)
        with open(assemble, mode="r", encoding="utf8") as file_handler:
            assemble = yaml.safe_load(file_handler)
        NamelistGeneratorAssemble.__init__(
            self, program, definitions, assemble, macros=macros, micro=micro
        )



"""Observation."""

import logging

import numpy as np


class Observation(object):
    """Observation class."""

    def __init__(
        self, obstime, lon, lat, value, elev=np.nan, stid="NA", varname=None, sigmao=1.0
    ):
        """Construct an observation.

        Args:
            obstime (as_datetime): Observation time
            lon (float): Longitude
            lat (float): Latitude
            value (float): Observation value
            elev (float, optional): Elevation. Defaults to np.nan.
            stid (str, optional): Staation ID. Defaults to "NA".
            varname (str, optional): _description_. Defaults to None.
            sigmao (float, optional): Observation error relative to a
                                      normalized background error (1). Defaults to 1.0

        """
        self.obstime = obstime
        self.lon = float(lon)
        self.lat = float(lat)
        self.posid = self.get_posid(self.lon, self.lat)
        self.stid = stid
        self.elev = float(elev)
        self.value = float(value)
        self.varname = varname
        self.sigmao = sigmao

    def print_obs(self):
        """Print observation."""
        logging.info(
            "observation: %s %s %s %s %s %s %s",
            self.obstime,
            self.lon,
            self.lat,
            self.stid,
            self.value,
            self.elev,
            self.sigmao,
        )

    @staticmethod
    def vectors2obs(obstime, lon, lat, stid, elev, value, varname, sigmao):
        """Convert vectors to observations.

        Args:
            obstime (_type_): _description_
            lon (_type_): _description_
            lat (_type_): _description_
            stid (_type_): _description_
            elev (_type_): _description_
            value (_type_): _description_
            varname (_type_): _description_
            sigmao(float): Observation error

        Returns:
            Observation: Observation object.
        """
        return Observation(
            obstime, lon, lat, value, elev=elev, varname=varname, stid=stid, sigmao=sigmao
        )

    @staticmethod
    def obs2vectors(my_obs):
        """Convert observations to vectors.

        Args:
            my_obs (Observation): Observation object

        Returns:
            tuple: _description_
        """
        return (
            my_obs.obstime,
            my_obs.lon,
            my_obs.lat,
            my_obs.stid,
            my_obs.elev,
            my_obs.value,
            my_obs.varname,
            my_obs.sigmao,
        )

    @staticmethod
    def format_lon(lon):
        """Format longitude."""
        lon = f"{float(lon):10.5f}"
        return lon

    @staticmethod
    def format_lat(lat):
        """Format latitude."""
        lat = f"{float(lat):10.5f}"
        return lat

    @staticmethod
    def get_posid(lon, lat, pos_decimals=5):
        """Get pos id.

        Args:
            lon (_type_): _description_
            lat (_type_): _description_
            pos_decimals (int, optional): _description_. Defaults to 5.

        Returns:
            _type_: _description_
        """
        return f"{lon:.{pos_decimals}f}#{lat:.{pos_decimals}f}"



"""Converter and read data."""

import copy
import logging
from abc import ABCMeta, abstractmethod

import numpy as np

from .cmd_parsing import (
    converter_parse_options,
    parse_args_variable,
    variable_parse_options,
)
from .datetime_utils import as_datetime
from .interpolation import fill_field
from .util import deep_update
from .variable import Variable


class ReadData(object):
    """Read data class."""

    __metaclass__ = ABCMeta

    def __init__(self, geo, var_name):
        """Construct readData object."""
        self.geo = geo
        self.var_name = var_name

    @abstractmethod
    def read_time_step(self, validtime, cache):
        """To be implemented."""
        raise NotImplementedError(
            "users must define read_time_step to use this base class"
        )

    @abstractmethod
    def print_info(self):
        """To be implemented."""
        raise NotImplementedError(
            "users must define read_time_step to use this base class"
        )


# Direct data can be ead with this class with converter = None
class ConvertedInput(ReadData):
    """Converted input.

    Args:
        geo (surfex.Geo): Surfex geometry
        var_name (str): Variable name
        converter (surfex.Converter): Converter
    """

    def __init__(self, geo, var_name, converter):
        """Construct converted input.

        Args:
            geo (_type_): _description_
            var_name (_type_): _description_
            converter (_type_): _description_

        """
        ReadData.__init__(self, geo, var_name)
        self.converter = converter

    def read_time_step(self, validtime, cache):
        """Read time  step.

        Args:
            validtime (_type_): _description_
            cache (_type_): _description_

        Returns:
            _type_: _description_

        """
        field = self.converter.read_time_step(self.geo, validtime, cache)
        return field

    def print_info(self):
        """Print info."""
        return self.converter.print_info()


class ConstantValue(ReadData):
    """Constant value converter."""

    def __init__(self, geo, var_name, var_dict):
        """Set a field to a constant value.

        Args:
            geo (surfex.Geo): Geometry object
            var_name (str): Variable name
            var_dict (dict): Variable definition

        Raises:
            RuntimeError: Constant value must have a value!

        """
        ReadData.__init__(self, geo, var_name)
        self.var_dict = var_dict
        if "value" in self.var_dict:
            self.value = self.var_dict["value"]
        else:
            raise RuntimeError("Constant value must have a value!")

    def read_time_step(self, validtime, cache):  # noqa ARG002
        """Read time step.

        Args:
            validtime (_type_): _description_
            cache (_type_): _description_

        Returns:
            _type_: _description_

        """
        field = np.array([float(i) for i in range(self.geo.npoints)])
        field.fill(self.value)
        return field

    def print_info(self):
        """Print info."""
        return self.var_name


#######################################################
#######################################################


class Converter(object):
    """Converter.

    Main interface to read a field is done through a converter.
    The converter name is default "None" to read a plain field without any conversion.

    """

    def __init__(self, name, initial_time, defs, conf, fileformat):
        """Initialize the converter.

        Args:
            name (str): Converter name
            initial_time (as_datetime): The valid time you want to read
            defs (dict): A dictionary defining the variables
            conf (dict): A dictionary defining the converter
            fileformat (str): File format

        Raises:
            KeyError: Missing definitions
            NotImplementedError: Converter not implemented

        """
        self.name = name
        self.initial_time = initial_time
        logging.info("initial time %s", self.initial_time)

        logging.debug("Converter name: %s", self.name)
        logging.debug("Converter config: %s", conf)
        if self.name not in conf:
            logging.debug("conf: %s", conf)
            raise KeyError(self.name + " is missing in converter definition")

        if self.name in ("none", "analysis"):
            self.var = self.create_variable(fileformat, defs, conf[self.name])
        elif name == "diff":
            self.field1 = self.create_variable(
                fileformat, defs, conf[self.name]["field1"]
            )
            self.field2 = self.create_variable(
                fileformat, defs, conf[self.name]["field2"]
            )
        elif name == "rh2q":
            self.r_h = self.create_variable(fileformat, defs, conf[self.name]["rh"])
            self.temp = self.create_variable(fileformat, defs, conf[self.name]["t"])
            self.pres = self.create_variable(fileformat, defs, conf[self.name]["p"])
        elif name == "mslp2ps":
            self.altitude = self.create_variable(
                fileformat, defs, conf[self.name]["altitude"]
            )
            self.temp = self.create_variable(fileformat, defs, conf[self.name]["t"])
            self.pres = self.create_variable(fileformat, defs, conf[self.name]["mslp"])
        elif name == "rh2q_mslp":
            self.r_h = self.create_variable(fileformat, defs, conf[self.name]["rh"])
            self.temp = self.create_variable(fileformat, defs, conf[self.name]["t"])
            self.altitude = self.create_variable(
                fileformat, defs, conf[self.name]["altitude"]
            )
            self.pres = self.create_variable(fileformat, defs, conf[self.name]["mslp"])
        elif name == "rh2q_z":
            self.r_h = self.create_variable(fileformat, defs, conf[self.name]["rh"])
            self.temp = self.create_variable(fileformat, defs, conf[self.name]["t"])
            self.altitude = self.create_variable(
                fileformat, defs, conf[self.name]["altitude"]
            )
        elif name == "td2q":
            self.td = self.create_variable(fileformat, defs, conf[self.name]["td"])
            self.pres = self.create_variable(fileformat, defs, conf[self.name]["p"])
        elif name == "td2rh":
            self.dewpoint = self.create_variable(fileformat, defs, conf[self.name]["td"])
            self.temperature = self.create_variable(
                fileformat, defs, conf[self.name]["t"]
            )
        elif name in ("windspeed", "winddir"):
            self.x_wind = self.create_variable(fileformat, defs, conf[self.name]["x"])
            self.y_wind = self.create_variable(fileformat, defs, conf[self.name]["y"])
        elif name in ("totalprec", "calcsnow"):
            self.totalprec = self.create_variable(
                fileformat, defs, conf[self.name]["totalprec"]
            )
            if name == "totalprec":
                self.snow = self.create_variable(
                    fileformat, defs, conf[self.name]["snow"]
                )
            elif name == "calcsnow":
                self.temp = self.create_variable(fileformat, defs, conf[self.name]["t"])
        elif name == "calcrain":
            self.totalprec = self.create_variable(
                fileformat, defs, conf[self.name]["totalprec"]
            )
            self.temp = self.create_variable(fileformat, defs, conf[self.name]["t"])
        elif name == "snowplusgraupel":
            self.snow = self.create_variable(fileformat, defs, conf[self.name]["snow"])
            self.graupel = self.create_variable(
                fileformat, defs, conf[self.name]["graupel"]
            )
        elif name == "phi2m":
            self.phi = self.create_variable(fileformat, defs, conf[self.name]["phi"])
        elif self.name in ("swe2sd", "sweclim"):
            self.swe = self.create_variable(fileformat, defs, conf[self.name]["swe"])
        elif self.name == "sea2land":
            self.sea = self.create_variable(fileformat, defs, conf[self.name]["sea"])
        elif self.name == "tap":
            self.tap1 = self.create_variable(fileformat, defs, conf[self.name]["tap1"])
            self.tap2 = self.create_variable(fileformat, defs, conf[self.name]["tap2"])
        elif self.name == "rhp":
            self.rhp1 = self.create_variable(fileformat, defs, conf[self.name]["rhp1"])
            self.rhp2 = self.create_variable(fileformat, defs, conf[self.name]["rhp2"])
        elif self.name == "sdp":
            self.sdp1 = self.create_variable(fileformat, defs, conf[self.name]["sdp1"])
            self.sdp2 = self.create_variable(fileformat, defs, conf[self.name]["sdp2"])
        elif self.name == "smp":
            self.smp1 = self.create_variable(fileformat, defs, conf[self.name]["smp1"])
            self.smp2 = self.create_variable(fileformat, defs, conf[self.name]["smp2"])
        elif self.name == "nature_town":
            self.nature = self.create_variable(
                fileformat, defs, conf[self.name]["nature"]
            )
            self.town = self.create_variable(fileformat, defs, conf[self.name]["town"])
        elif self.name == "cloud_base":
            self.cloud_base = self.create_variable(
                fileformat, defs, conf[self.name]["cloud_base"]
            )
        else:
            raise NotImplementedError("Converter " + self.name + " not implemented")

        logging.debug("Constructed the converter %s", self.name)

    def print_info(self):
        """Print info."""
        return f"name: {self.name}"

    def create_variable(self, fileformat, defs, var_dict):
        """Create a variable.

        Args:
            fileformat (str): Fileformat
            defs (dict): defs
            var_dict (dict): Variable dictionary

        Raises:
            RuntimeError: Variable is not set

        Returns:
            field: The read field

        """
        # Finally we can merge the variable with the default settings
        # Create deep copies not to inherit between variables
        defs = copy.deepcopy(defs)
        var_dict = copy.deepcopy(var_dict)
        if var_dict is None:
            raise RuntimeError("Variable is not set")
        merged_dict = deep_update(defs, var_dict)

        var = Variable(fileformat, merged_dict, self.initial_time)

        logging.debug(var.print_variable_info())
        return var

    @staticmethod
    def mslp2ps(mslp, altitude, temp):
        """Calcaulate ps from mslp.

        Args:
            mslp (np.ndarray): Mean sea level pressure
            altitude (np.ndarray): Altitude
            temp (np.ndarray): Temperature

        Returns:
            np.ndarray: Surface pressure

        """
        gravity = 9.81
        dry_air = 287.0

        pres = np.multiply(
            mslp,
            np.exp(
                np.divide(np.multiply(-altitude, gravity), np.multiply(dry_air, temp))
            ),
        )

        return pres

    @staticmethod
    def saturation_mixing_ratio(total_press, temperature):
        """Calculate sat mixing ratio.

        Args:
            total_press (np.ndarray): Pressure
            temperature (np.ndarray): Temperature

        Returns:
            (np.ndarray): saturation mixing ratio

        """
        return Converter.mixing_ratio(
            Converter.saturation_vapor_pressure(temperature), total_press
        )

    @staticmethod
    def mixing_ratio(partial_press, total_press, molecular_weight_ratio=0.622):
        """Calculate mixing ratio.

        Args:
            partial_press (np.ndarray): Partial pressure
            total_press (np.ndarray): Total pressure
            molecular_weight_ratio (float, optional): Molecular weight ratio.
                                                      Defaults to 0.622

        Returns:
            (np.ndarray): mixing ratio

        """
        return np.multiply(
            molecular_weight_ratio,
            np.divide(partial_press, np.subtract(total_press, partial_press)),
        )

    @staticmethod
    def specific_humidity_from_dewpoint(pressure, dewpoint):
        """Calculate q from Td.

        Args:
            pressure (np.ndarray): Pressure
            dewpoint (np.ndarray): Temperature

        Returns:
            (np.ndarray): specific humidity

        """
        mixing_ratio = Converter.saturation_mixing_ratio(pressure, dewpoint)
        return Converter.specific_humidity_from_mixing_ratio(mixing_ratio)

    @staticmethod
    def mixing_ratio_from_specific_humidity(specific_humidity):
        """Calculate mixing ratio from q.

        Args:
            specific_humidity (np.ndarray): Specific humidity

        Returns:
            (np.ndarray): mixing ratio

        """
        return np.divide(specific_humidity, np.subtract(1, specific_humidity))

    @staticmethod
    def specific_humidity_from_mixing_ratio(mixing_ratio):
        """Calculate q from mixing ratio.

        Args:
            mixing_ratio (np.ndarray): Mixing ratio

        Returns:
            (np.ndarray): Specific humidity

        """
        return np.divide(mixing_ratio, np.subtract(1, mixing_ratio))

    @staticmethod
    def saturation_vapor_pressure(temperature_kelvin):
        """Calculate esat.

        Args:
            temperature_kelvin (np.ndarray): Absolute temperature

        Returns:
            (np.ndarray): stauration vapor pressure

        """
        field_t_c = np.subtract(temperature_kelvin, 273.15)
        exp = np.divide(np.multiply(17.67, field_t_c), np.add(field_t_c, 243.5))
        esat = np.multiply(6.112, np.exp(exp))
        return esat

    def read_time_step(self, geo, validtime, cache):
        """Read time step.

        Args:
            geo (Geo): Geometry_
            validtime (as_datetime): Validtime
            cache (Cache): Cache

        Raises:
            KeyError:Could not found climatological mean for month
            NotImplementedError: _description_
            RuntimeError: "Alpha is different for the 2 wind vectors!

        Returns:
            field (np.ndarray): Read and converted field

        """
        gravity = 9.81
        field = None
        # Specific reading for each converter
        if self.name in ("none", "analysis"):
            field = self.var.read_variable(geo, validtime, cache)
        elif self.name == "diff":
            field = self.field1.read_variable(
                geo, validtime, cache
            ) - self.field2.read_variable(geo, validtime, cache)
        elif self.name in ("windspeed", "winddir"):
            field_x = self.x_wind.read_variable(geo, validtime, cache)
            field_y = self.y_wind.read_variable(geo, validtime, cache)
            if self.name == "windspeed":
                field = np.sqrt(np.square(field_x) + np.square(field_y))
                np.where(field < 0.005, field, 0)
            elif self.name == "winddir":
                field = np.mod(180 + np.rad2deg(np.arctan2(field_x, field_y)), 360)
                if self.x_wind.alpha is not None and self.y_wind.alpha is not None:
                    if self.x_wind.alpha.all() == self.y_wind.alpha.all():
                        field = np.subtract(field, self.x_wind.alpha)
                        logging.debug("Wind was rotated to geographical coordinates")
                    else:
                        raise RuntimeError("Alpha is different for the 2 wind vectors!")
                else:
                    logging.warning(
                        "Wind was not rotated to geographical coordinates due to %s",
                        "missing alphas",
                    )

        elif self.name in ("rh2q", "rh2q_mslp", "rh2q_z"):
            field_r_h = self.r_h.read_variable(geo, validtime, cache)  # %
            field_temp = self.temp.read_variable(geo, validtime, cache)  # In K
            if self.name == "rh2q_z":
                field_pres = np.array(field_r_h.shape)
                field_pres.fill(101325.0)
            else:
                field_pres = self.pres.read_variable(geo, validtime, cache)  # In Pa
            if self.name in ("rh2q_mslp", "rh2q_z"):
                field_altitude = self.altitude.read_variable(
                    geo, validtime, cache
                )  # In m
                field_pres = self.mslp2ps(field_pres, field_altitude, field_temp)

            field_p_mb = np.divide(field_pres, 100.0)
            field_t_c = np.subtract(field_temp, 273.15)

            exp = np.divide(np.multiply(17.67, field_t_c), np.add(field_t_c, 243.5))
            esat = np.multiply(6.112, np.exp(exp))
            field = np.divide(np.multiply(0.622, field_r_h / 100.0) * esat, field_p_mb)

        elif self.name == "td2rh":
            dewpoint = self.dewpoint.read_variable(geo, validtime, cache)  # In K
            temperature = self.temperature.read_variable(geo, validtime, cache)  # In K
            td_e = Converter.saturation_vapor_pressure(dewpoint)
            t_e = Converter.saturation_vapor_pressure(temperature)
            field = np.divide(td_e, t_e)

        elif self.name in ("td2q", "td2q_z", "td2q_mslp"):
            field_td = self.td.read_variable(geo, validtime, cache)  # %
            if self.name == "td2q_z":
                field_pres = np.array(field_td.shape)
                field_pres.fill(101325.0)
            else:
                field_pres = self.pres.read_variable(geo, validtime, cache)  # In Pa
            if self.name in ("td2q_mslp", "td2q_z"):
                field_altitude = self.altitude.read_variable(
                    geo, validtime, cache
                )  # In m
                field_pres = self.mslp2ps(field_pres, field_altitude, field_td)

            field_p_mb = np.divide(field_pres, 100.0)
            field = Converter.specific_humidity_from_dewpoint(field_p_mb, field_td)

        elif self.name == "mslp2ps":
            field_pres = self.pres.read_variable(geo, validtime, cache)  # In Pa
            field_temp = self.temp.read_variable(geo, validtime, cache)  # In K
            field_altitude = self.altitude.read_variable(geo, validtime, cache)  # In m
            field = self.mslp2ps(field_pres, field_altitude, field_temp)
        elif self.name == "totalprec":
            field_totalprec = self.totalprec.read_variable(geo, validtime, cache)
            field_snow = self.snow.read_variable(geo, validtime, cache)
            field = np.subtract(field_totalprec, field_snow)
            if any(field[field < 0.0]):
                logging.info("Set negative rain values to zero")
                field[field < 0.0] = 0

        elif self.name == "calcrain":
            field_totalprec = self.totalprec.read_variable(geo, validtime, cache)
            field_t = self.temp.read_variable(geo, validtime, cache)
            field = field_totalprec
            field[field_t <= 274.16] = 0
        elif self.name == "calcsnow":
            field_totalprec = self.totalprec.read_variable(geo, validtime, cache)
            field_t = self.temp.read_variable(geo, validtime, cache)  # In K
            field = field_totalprec
            field[field_t > 274.16] = 0
        elif self.name == "snowplusgraupel":
            field = self.snow.read_variable(geo, validtime, cache)
            field += self.graupel.read_variable(geo, validtime, cache)
        elif self.name == "phi2m":
            field = self.phi.read_variable(geo, validtime, cache)
            field = np.divide(field, gravity)
            field[(field < 0)] = 0.0
        elif self.name == "swe2sd":
            field = self.swe.read_variable(geo, validtime, cache)
            rho = self.swe.read_variable(geo, validtime, cache)
            field = np.divide(field, rho)
        elif self.name == "sweclim":
            field = self.swe.read_variable(geo, validtime, cache)
            rhoclim = {
                "01": 222.0,
                "02": 233.0,
                "03": 240.0,
                "04": 278.0,
                "05": 212.0,
                "06": 312.0,
                "07": 312.0,
                "08": 143.0,
                "09": 143.0,
                "10": 161.0,
                "11": 182.0,
                "12": 213.0,
            }
            month = validtime.strftime("%m")
            if month in rhoclim:
                field = np.divide(field, rhoclim[month])
            else:
                raise KeyError(
                    "Could not found climatological mean for month " + str(month)
                )
        elif self.name == "sea2land":
            field = self.sea.read_variable(geo, validtime, cache)
            field = np.subtract(1, field)
        elif self.name == "tap":
            tap1 = self.tap1.read_variable(geo, validtime, cache)
            tap2 = self.tap2.read_variable(geo, validtime, cache)
            field = np.where(np.isnan(tap1), tap2, tap1)
        elif self.name == "rhp":
            rhp1 = self.rhp1.read_variable(geo, validtime, cache)
            rhp2 = self.rhp2.read_variable(geo, validtime, cache)
            field = np.where(np.isnan(rhp1), rhp2, rhp1)
        elif self.name == "sdp":
            sdp1 = self.sdp1.read_variable(geo, validtime, cache)
            sdp2 = self.sdp2.read_variable(geo, validtime, cache)
            field = np.where(np.isnan(sdp1), sdp2, sdp1)
        elif self.name == "smp":
            smp1 = self.smp1.read_variable(geo, validtime, cache)
            smp2 = self.smp2.read_variable(geo, validtime, cache)
            field = np.where(np.isnan(smp1), smp2, smp1)
        elif self.name == "nature_town":
            nature = self.nature.read_variable(geo, validtime, cache)
            town = self.town.read_variable(geo, validtime, cache)
            field = np.add(nature, town)
            field[field > 1] = 1.0
        elif self.name == "cloud_base":
            logging.info("Converter cloud_base")

            field = self.cloud_base.read_variable(geo, validtime, cache)
            field_2d = field.reshape(geo.nlons, geo.nlats)

            iteration = 0
            while np.any(np.isnan(field_2d)):
                logging.debug("Filling cloud base")
                field_2d, nans = fill_field(field_2d, geo, radius=3)
                iteration = iteration + 1
                logging.debug("Iteration %s NaNs: %s", iteration, nans)

            # Reshape back to 1D
            field = field_2d.reshape(geo.nlons * geo.nlats)

        else:
            raise NotImplementedError(f"Converter {self.name} not implemented")
        return field


def kwargs2converter(**kwargs):
    """Create a converter object from keyword arguments.

    Args:
        kwargs(dict): keyword arguments

    Raises:
        RuntimeError: _description_
        RuntimeError: _description_
        RuntimeError: _description_
        RuntimeError: _description_
        RuntimeError: _description_
        RuntimeError: _description_
        RuntimeError: _description_
        RuntimeError: _description_
        RuntimeError: _description_
        RuntimeError: _description_
        NotImplementedError: _description_

    Returns:
        Converter: A converter object

    """
    try:
        converter_name = kwargs["converter"]
    except KeyError:
        converter_name = "none"
    try:
        conv_variables = kwargs["conv_variables"]
    except KeyError:
        conv_variables = None
    if conv_variables is None:
        conv_variables = ["var"]
    try:
        system_file_paths = kwargs["var"]["system_file_paths"]
    except KeyError:
        try:
            system_file_paths = kwargs["system_file_paths"]
        except KeyError:
            raise RuntimeError from KeyError

    var_dict2 = {}
    for conv_var in conv_variables:
        kwargs2 = kwargs[conv_var]

        prefer_forecast = False
        if kwargs2["preference"] == "forecast":
            prefer_forecast = True
        try:
            basetime = kwargs2["basetime"]
        except KeyError:
            basetime = None
        if basetime is not None and isinstance(basetime, str):
            basetime = as_datetime(kwargs2["basetime"])
        variable = None
        if "varname" in kwargs2:
            variable = kwargs2["varname"]
        filepattern = None
        if "filepattern" in kwargs2:
            filepattern = kwargs2["filepattern"]
        try:
            inputtype = kwargs2["inputtype"]
        except KeyError as exc:
            raise RuntimeError("Input type must be set") from exc

        try:
            interpolator = kwargs2["interpolator"]
        except KeyError:
            interpolator = "nearest"
        if inputtype == "grib1":
            if filepattern is None:
                raise RuntimeError("You must provide a filepattern")

            par = kwargs2["parameter"]
            ltp = kwargs2["levelType"]
            lev = kwargs2["level"]
            tri = kwargs2["timeRangeIndicator"]

            var_dict = {
                "filepattern": filepattern,
                "fcint": kwargs2["fcint"],
                "file_inc": kwargs2["file_inc"],
                "offset": kwargs2["offset"],
                "prefer_forecast": prefer_forecast,
                "parameter": par,
                "type": ltp,
                "level": lev,
                "tri": tri,
                "interpolator": interpolator,
                "system_file_paths": system_file_paths,
            }

        elif inputtype == "grib2":
            if filepattern is None:
                raise RuntimeError("You must provide a filepattern")

            discipline = kwargs2["discipline"]
            parameter_category = kwargs2["parameterCategory"]
            parameter_number = kwargs2["parameterNumber"]
            level_type = kwargs2["levelType"]
            level = kwargs2["level"]
            type_of_statistical_processing = kwargs2["typeOfStatisticalProcessing"]

            var_dict = {
                "fcint": kwargs2["fcint"],
                "file_inc": kwargs2["file_inc"],
                "offset": kwargs2["offset"],
                "prefer_forecast": prefer_forecast,
                "filepattern": filepattern,
                "discipline": discipline,
                "parameterCategory": parameter_category,
                "parameterNumber": parameter_number,
                "levelType": level_type,
                "level": level,
                "typeOfStatisticalProcessing": type_of_statistical_processing,
                "system_file_paths": system_file_paths,
            }

        elif inputtype in ("netcdf", "fa"):
            if variable is None:
                raise RuntimeError("You must provide a variable")
            if filepattern is None:
                raise RuntimeError("You must provide a filepattern")

            var_dict = {
                "name": variable,
                "filepattern": filepattern,
                "fcint": kwargs2["fcint"],
                "file_inc": kwargs2["file_inc"],
                "offset": kwargs2["offset"],
                "prefer_forecast": prefer_forecast,
                "interpolator": interpolator,
                "system_file_paths": system_file_paths,
            }
        elif inputtype == "surfex":
            if variable is None:
                raise RuntimeError("You must provide a variable")
            if filepattern is None:
                raise RuntimeError("You must provide a filepattern")
            try:
                patches = kwargs2["sfx_patches"]
            except KeyError:
                patches = None
            try:
                layers = kwargs2["sfx_layers"]
            except KeyError:
                layers = None
            try:
                datatype = kwargs2["sfx_datatype"]
            except KeyError:
                datatype = None
            try:
                interval = kwargs2["sfx_interval"]
            except KeyError:
                interval = None
            try:
                geo_sfx_input = kwargs2["sfx_geo_input"]
            except KeyError:
                geo_sfx_input = None

            var_dict = {
                "varname": variable,
                "filepattern": filepattern,
                "patches": patches,
                "layers": layers,
                "datatype": datatype,
                "interval": interval,
                "basetime": basetime,
                "fcint": kwargs2["fcint"],
                "file_inc": kwargs2["file_inc"],
                "offset": kwargs2["offset"],
                "prefer_forecast": prefer_forecast,
                "interpolator": interpolator,
                "system_file_paths": system_file_paths,
            }
            if geo_sfx_input is not None:
                var_dict.update({"geo_input_file": geo_sfx_input})

        elif inputtype == "obs":
            if variable is None:
                raise RuntimeError("You must provide a variable")

            obs_input_type = kwargs["var"]["obs_type"]
            if obs_input_type is None:
                raise RuntimeError("You must provide an obs type")

            var_dict = {
                "filetype": obs_input_type,
                "varname": [variable],
                "filepattern": filepattern,
                "filenames": [filepattern],
                "fcint": 10800,
                "file_inc": 10800,
                "offset": 0,
                "system_file_paths": system_file_paths,
            }
        else:
            raise NotImplementedError
        if converter_name != "none":
            var_dict2.update({conv_var: var_dict})
        else:
            var_dict2 = var_dict

    defs = {variable: {inputtype: {"converter": {converter_name: var_dict2}}}}

    converter_conf = defs[variable][inputtype]["converter"]
    return Converter(converter_name, basetime, defs, converter_conf, inputtype)


def get_multi_converters(parser, multivars, argv):
    """Get multi converter.

    Args:
        parser (Parser): Argument parser
        multivars (list): The variables
        argv (list): Argument list

    Returns:
        converters (dict): A dictionary with converters if several variables
                           or a converter object

    """
    single = False
    if multivars is None or len(multivars) == 0:
        multivars = ["var"]
        single = True
    for mvar in multivars:
        prefix = mvar
        if single:
            prefix = ""
        converter_parse_options(parser, prefix=prefix)

    args, __ = parser.parse_known_args(argv)
    kwargs = {}
    for arg in vars(args):
        logging.debug("arg=%s val=%s", arg, getattr(args, arg))
        kwargs.update({arg: getattr(args, arg)})

    #################################################

    # Set converter input
    converters = {}
    for mvar in multivars:
        prefix = f"{mvar}_"
        if single:
            prefix = ""
        converters.update(
            {
                mvar: {
                    "converter": kwargs[f"{prefix}converter"],
                    "variables": kwargs[f"{prefix}conv_variables"],
                }
            }
        )

    # Set converter options
    for mvar in multivars:
        variables = converters[mvar]["variables"]
        if variables is None:
            var_prefix = f"{mvar}"
            if single:
                var_prefix = ""
            variable_parse_options(parser, name=var_prefix)
        else:
            for var in variables:
                var_prefix = f"{mvar}-{var}"
                if single:
                    var_prefix = f"{var}"
                variable_parse_options(parser, name=var_prefix)

    converters2 = {}
    for mvar in multivars:
        variables = converters[mvar]["variables"]
        logging.info("%s %s", mvar, variables)
        if variables is None:
            prefix = f"{mvar}"
            if single:
                prefix = ""
            kwargs = parse_args_variable(parser, {}, argv, prefix=prefix)
        else:
            prefix = f"{mvar}"
            if single:
                prefix = ""
            kwargs = parse_args_variable(
                parser, {}, argv, variables=variables, prefix=prefix
            )
        obj = kwargs2converter(**kwargs)
        converters[mvar].update({"obj": obj})
        converters2.update({mvar: obj})
    if single:
        return converters2["var"]
    return converters2



"""Grib treatment."""

import logging

import numpy as np
import pyproj

try:
    import eccodes
    import gribapi
except ImportError:
    eccodes = None
    gribapi = None
except RuntimeError:
    eccodes = None
    gribapi = None
# Needed in Python 3.5
except:  # noqa
    eccodes = None
    gribapi = None


from .geo import ConfProj, Geo, LonLatReg
from .interpolation import Interpolation


class Grib(object):
    """Grib class."""

    def __init__(self, fname):
        """Construct grib object.

        Args:
            fname (str): File name.

        """
        self.fname = fname
        self.projection = None
        self.lons = None
        self.lats = None
        self.nearest = None
        self.linear = None
        logging.debug("Grib constructor")

    def field(self, gribvar, time):
        """Read field in grib file.

        Args:
            gribvar (GribVariable1/2): Grib variable to read.
            time (datetime.datetime): Valid time to read.

        Returns:
            np.ndarray: Field

        Raises:
            NotImplementedError: NotImplementedError
            RuntimeError: If eccodes not available

        """
        if eccodes is None:
            raise RuntimeError("eccodes not found. Needed for reading grib files")

        logging.debug("Look for %s", gribvar.generate_grib_id())

        field = None
        geo_out = None
        with open(self.fname, mode="rb") as file_handler:
            while 1:
                gid = eccodes.codes_grib_new_from_file(file_handler)

                if gid is None:
                    logging.warning("Could not find key")
                    gribvar.print_keys()
                    file_handler.close()
                    return field, geo_out

                if gribvar.matches(gid):
                    values = self.read_field_in_message(gid, time)
                    logging.debug("read values = %s", values)

                    grid_type = str(eccodes.codes_get(gid, "gridType"))
                    logging.debug("grid_type=%s", grid_type)
                    if grid_type.lower() == "rotated_ll":
                        geo_keys = [
                            "Ni",
                            "Nj",
                            "latitudeOfFirstGridPointInDegrees",
                            "longitudeOfFirstGridPointInDegrees",
                            "latitudeOfLastGridPointInDegrees",
                            "longitudeOfLastGridPointInDegrees",
                            "iDirectionIncrementInDegrees",
                            "jDirectionIncrementInDegrees",
                            "latitudeOfSouthernPoleInDegrees",
                            "longitudeOfSouthernPoleInDegrees",
                            "iScansNegatively",
                            "jScansPositively",
                        ]
                        geo_info = self.read_geo_info(gid, geo_keys)

                        n_x = geo_info["Ni"]
                        n_y = geo_info["Nj"]

                        ll_lon = geo_info["longitudeOfFirstGridPointInDegrees"]
                        ll_lat = geo_info["latitudeOfFirstGridPointInDegrees"]
                        dlon = geo_info["iDirectionIncrementInDegrees"]
                        dlat = geo_info["jDirectionIncrementInDegrees"]
                        sp_lon = geo_info["longitudeOfSouthernPoleInDegrees"]
                        iscan = geo_info["iScansNegatively"]
                        jscan = geo_info["jScansPositively"]
                        if sp_lon < -180.0:
                            sp_lon = sp_lon + 360.0
                        elif sp_lon > 180.0:
                            sp_lon = sp_lon - 360.0
                        sp_lat = -1 * geo_info["latitudeOfSouthernPoleInDegrees"]
                        earth = 6.371229e6

                        proj_string = (
                            f"+proj=ob_tran +o_proj=longlat +o_lat_p={sp_lat}"
                            f" +R={earth!s} +no_defs"
                        )
                        logging.info(proj_string)
                        logging.info("ll_lon=%s ll_lat=%s", ll_lon, ll_lat)
                        logging.info("polon=%s polat=%s", sp_lon, sp_lat)
                        logging.info("dlon=%s dlat=%s", dlon, dlat)
                        logging.info("iscan=%s jscan=%s", iscan, jscan)
                        proj = pyproj.CRS.from_string(proj_string)
                        wgs84 = pyproj.CRS.from_string("EPSG:4326")

                        lons = []
                        for i in range(n_x):
                            if int(iscan) == 1:
                                lon = ll_lon - (float(i) * dlon)
                            else:
                                lon = ll_lon + (float(i) * dlon)
                            if lon < -180.0:
                                lon = lon + 360.0
                            elif lon > 180.0:
                                lon = lon - 360.0
                            lons.append(lon)
                        lats = []
                        for j in range(n_y):
                            if int(jscan) == 1:
                                lat = ll_lat + (float(j) * dlat)
                            else:
                                lat = ll_lat - (float(j) * dlat)
                            if lat > 90.0:
                                lat = lat - 90.0
                            elif lat < -90.0:
                                lat = lat + 90.0
                            lats.append(lat)

                        lons = np.array(lons)
                        lats = np.array(lats)
                        longitudes, latitudes = np.meshgrid(lons, lats, indexing="ij")
                        lons, lats = pyproj.Transformer.from_crs(
                            proj, wgs84, always_xy=True
                        ).transform(longitudes, latitudes)
                        lons = lons + sp_lon

                        field = np.reshape(values, [n_x, n_y], order="F")
                        if geo_out is None:
                            geo_out = Geo(lons, lats)

                    elif grid_type.lower() == "regular_ll":
                        geo_keys = [
                            "Ni",
                            "Nj",
                            "latitudeOfFirstGridPointInDegrees",
                            "longitudeOfFirstGridPointInDegrees",
                            "latitudeOfLastGridPointInDegrees",
                            "longitudeOfLastGridPointInDegrees",
                            "iDirectionIncrementInDegrees",
                            "jDirectionIncrementInDegrees",
                        ]
                        geo_info = self.read_geo_info(gid, geo_keys)
                        n_x = geo_info["Ni"]
                        n_y = geo_info["Nj"]
                        lon0 = geo_info["longitudeOfFirstGridPointInDegrees"]
                        lat0 = geo_info["latitudeOfFirstGridPointInDegrees"]
                        d_x = geo_info["iDirectionIncrementInDegrees"]
                        d_y = geo_info["jDirectionIncrementInDegrees"]
                        lons = []
                        lats = []
                        for i in range(n_x):
                            lons.append(lon0 + (float(i) * d_x))
                        for j in range(n_y):
                            lats.append(lat0 - (float(j) * d_y))
                        lon1 = lons[-1]
                        lat1 = lats[-1]
                        lons = np.array(lons)
                        lats = np.array(lats)
                        lons, lats = np.meshgrid(lons, lats)
                        field = np.reshape(values, [n_x, n_y], order="F")

                        if geo_out is None:
                            domain = {
                                "nam_lonlat_reg": {
                                    "xlonmin": lon0,
                                    "xlonmax": lon1,
                                    "xlatmin": lat0,
                                    "xlatmax": lat1,
                                    "nlon": n_x,
                                    "nlat": n_y,
                                }
                            }
                            geo_out = LonLatReg(domain)

                    elif grid_type.lower() == "lambert":
                        geo_keys = [
                            "Nx",
                            "Ny",
                            "latitudeOfFirstGridPointInDegrees",
                            "longitudeOfFirstGridPointInDegrees",
                            "LoVInDegrees",
                            "DxInMetres",
                            "DyInMetres",
                            "iScansNegatively",
                            "jScansPositively",
                            "jPointsAreConsecutive",
                            "Latin1InDegrees",
                            "LaDInDegrees",
                            "Latin2InDegrees",
                            "latitudeOfSouthernPoleInDegrees",
                            "longitudeOfSouthernPoleInDegrees",
                        ]
                        geo_info = self.read_geo_info(gid, geo_keys)

                        n_x = geo_info["Nx"]
                        n_y = geo_info["Ny"]

                        lon0 = geo_info["LoVInDegrees"]
                        lat0 = geo_info["LaDInDegrees"]
                        ll_lon = geo_info["longitudeOfFirstGridPointInDegrees"]
                        ll_lat = geo_info["latitudeOfFirstGridPointInDegrees"]
                        d_x = geo_info["DxInMetres"]
                        d_y = geo_info["DyInMetres"]

                        earth = 6.37122e6
                        proj_string = (
                            f"+proj=lcc +lat_0={lat0!s} +lon_0={lon0!s} "
                            f"+lat_1={lat0!s} +lat_2={lat0!s} "
                            f"+units=m +no_defs +R={earth!s}"
                        )

                        proj = pyproj.CRS.from_string(proj_string)
                        wgs84 = pyproj.CRS.from_string("EPSG:4326")
                        x_0, y_0 = pyproj.Transformer.from_crs(
                            wgs84, proj, always_xy=True
                        ).transform(ll_lon, ll_lat)
                        x_c = x_0 + 0.5 * (n_x - 1) * d_x
                        y_c = y_0 + 0.5 * (n_y - 1) * d_y
                        lonc, latc = pyproj.Transformer.from_crs(
                            proj, wgs84, always_xy=True
                        ).transform(x_c, y_c)

                        # TODO we should investigate scan angle and if done correctly
                        # order should probaly be "C" and not "F"
                        field = np.reshape(values, [n_x, n_y], order="F")

                        if geo_out is None:
                            domain = {
                                "nam_conf_proj": {"xlon0": lon0, "xlat0": lat0},
                                "nam_conf_proj_grid": {
                                    "xloncen": lonc,
                                    "xlatcen": latc,
                                    "nimax": n_x,
                                    "njmax": n_y,
                                    "xdx": d_x,
                                    "xdy": d_y,
                                    "ilone": 0,
                                    "ilate": 0,
                                },
                            }
                            geo_out = ConfProj(domain)
                    else:
                        raise NotImplementedError(
                            str(grid_type) + " not implemented yet!"
                        )

                    eccodes.codes_release(gid)
                    if geo_out is None:
                        raise RuntimeError("No geometry is found in file")

                    return field, geo_out
                eccodes.codes_release(gid)
            raise RuntimeError("This should not happen")

    @staticmethod
    def read_geo_info(gid, keys):
        """Set geo keys in dict.

        Args:
            gid (int): grib id
            keys (dict): Geometry keys

        Returns:
            dict: Geometry

        """
        geo_dict = {}
        for key in keys:
            try:
                logging.debug("  %s: %s", key, eccodes.codes_get(gid, key))
                geo_dict.update({key: eccodes.codes_get(gid, key)})
            except eccodes.KeyValueNotFoundError as err:  # noqa PERF203
                logging.debug('  Key="%s" was not found: %s', key, err.msg)
            except eccodes.CodesInternalError as err:
                logging.error('Error with key="%s" : %s', key, err.msg)
        return geo_dict

    @staticmethod
    def read_field_in_message(gid, time):
        """Get field.

        Args:
            gid (int): grib id
            time(str): time

        Returns:
            Field: Field object

        """
        try:
            logging.debug(
                "There are %d values, average is %f, min is %f, max is %f",
                eccodes.codes_get_size(gid, "values"),
                eccodes.codes_get(gid, "average"),
                eccodes.codes_get(gid, "min"),
                eccodes.codes_get(gid, "max"),
            )
            field = eccodes.codes_get_values(gid)

            # TODO Check time consistency
            logging.info("Grib record found is hopefullly valid for time %s", str(time))

            has_bitmap = 0
            try:
                has_bitmap = int(eccodes.codes_get(gid, "bitmapPresent"))
            except eccodes.KeyValueNotFoundError as err:
                logging.debug('  Key="bitmapPresent" was not found: %s', err.msg)
            except eccodes.CodesInternalError as err:
                logging.error('Error with key="bitmapPresent" : %s', err.msg)
            if has_bitmap == 1:
                missing_value = eccodes.codes_get(gid, "missingValue")
                field[field == missing_value] = np.nan
            return field
        except eccodes.KeyValueNotFoundError as err:
            logging.debug('  Key="missingValue" was not found: %s', err.msg)
        except eccodes.CodesInternalError as err:
            logging.error('Error with key="missingValue" : %s', err.msg)
        return None

    def points(self, gribvar, geo, validtime=None, interpolation="bilinear"):
        """Read a 2-D field and interpolates it to requested positions.

        Args:
            gribvar (GribVariable1/2): Grib variable
            geo (surfex.Geo): Surfex geometry
            validtime (datetime.datetime, optional): Valid time. Defaults to None.
            interpolation (str, optional): Interpolation method. Defaults to "bilinear".

        Returns:
             np.array: vector with interpolated values

        """
        field, geo_in = self.field(gribvar, validtime)
        interpolator = Interpolation(interpolation, geo_in, geo)
        field = interpolator.interpolate(field)
        return field, interpolator


class Grib1Variable:
    """Grib1 variable."""

    def __init__(self, par, typ, level, tri=0):
        """Construct grib1 variable.

        Args:
            par (int): IndicatorOfParameter
            typ (int): TypeOfLevel
            level (int): NumberOFLevel
            tri (int, optional): TimeRangeIndicator. Defaults to 0.

        """
        self.version = 1
        self.par = int(par)
        self.typ = int(typ)
        self.level = int(level)
        self.tri = int(tri)

    def is_accumulated(self):
        """Check if accumulated."""
        if self.tri == 4:
            logging.debug("Value is accumulated")
            return True
        return False

    def matches(self, gid):
        """Check if matchees.

        Args:
            gid (_type_): _description_

        Returns:
            bool: True if found

        Raises:
            RuntimeError: If eccodes not found

        """
        if eccodes is None:
            raise RuntimeError("eccodes not found. Needed for reading grib files")

        version = int(eccodes.codes_get(gid, "editionNumber"))
        if version == 1:
            par = int(eccodes.codes_get_long(gid, "indicatorOfParameter"))
            lev = int(eccodes.codes_get_long(gid, "level"))
            typ = int(eccodes.codes_get_long(gid, "levelType"))
            tri = int(eccodes.codes_get_long(gid, "timeRangeIndicator"))

            logging.debug("Checking grib1 record: %s %s %s %s", par, typ, lev, tri)
            logging.debug(self.generate_grib_id())
            if (
                self.par == par
                and self.level == lev
                and self.typ == typ
                and self.tri == tri
            ):
                logging.debug(
                    "Found matching grib1 record: %s %s %s %s", par, lev, typ, tri
                )
                return True
            return False
        logging.warning("Record is not grib1")
        return False

    def print_keys(self):
        """Print keys."""
        logging.info("Version: %s", self.version)
        logging.info("indicatorOfParameter: %s", self.par)
        logging.info("levelType %s", self.typ)
        logging.info("level: %s", self.level)
        logging.info("timeRangeIndicator: %s", self.tri)

    def generate_grib_id(self):
        """Generate grib1 ID."""
        par = self.par
        typ = self.typ
        level = self.level
        tri = self.tri
        return f":grib1:{par!s}:{typ!s}:{level!s}:{tri!s}"


class Grib2Variable(object):
    """Grib2 variable."""

    def __init__(self, discipline, pca, pnr, typ, lev, tsp=-1):
        """Construct a grib2 variable.

        Args:
            discipline (int): discipline
            pca (int): parameterCatergory
            pnr (int): parameterNumber
            typ (int): levelType
            lev (int): level
            tsp (int, optional): typeOfStatisticalProcessing. Defaults to -1.
        """
        self.version = 2
        self.discipline = int(discipline)
        self.parameter_category = int(pca)
        self.parameter_number = int(pnr)
        self.level_type = int(typ)
        self.level = int(lev)
        self.type_of_statistical_processing = int(tsp)

    def matches(self, gid):
        """Check if matches.

        Args:
            gid (int): Grib ID

        Raises:
            ModuleNotFoundError: If not eccodes found

        Returns:
            bool: True if matches
        """
        if eccodes is None:
            raise ModuleNotFoundError("eccodes not found. Needed for reading grib files")

        version = int(eccodes.codes_get(gid, "editionNumber"))
        if version == 2:
            discipline = int(eccodes.codes_get(gid, "discipline"))
            parameter_category = int(eccodes.codes_get(gid, "parameterCategory"))
            parameter_number = int(eccodes.codes_get(gid, "parameterNumber"))
            level_type = int(eccodes.codes_get_long(gid, "levelType"))
            level = int(eccodes.codes_get(gid, "level"))
            try:
                type_of_statistical_processing = eccodes.codes_get(
                    gid, "typeOfStatisticalProcessing"
                )
                type_of_statistical_processing = int(type_of_statistical_processing)
            except gribapi.errors.KeyValueNotFoundError:
                type_of_statistical_processing = -1

            logging.debug(
                "Checking grib2 record: %s %s %s %s %s %s",
                discipline,
                parameter_category,
                parameter_number,
                level_type,
                level,
                type_of_statistical_processing,
            )
            logging.debug("grib2 ID: %s", self.generate_grib_id())
            if (
                self.discipline == discipline
                and self.parameter_category == parameter_category
                and self.parameter_number == parameter_number
                and self.level_type == level_type
                and self.level == level
                and self.type_of_statistical_processing == type_of_statistical_processing
            ):
                logging.debug(
                    "Found matching grib2 record: %s %s %s %s %s",
                    discipline,
                    parameter_category,
                    parameter_number,
                    level_type,
                    level,
                )
                return True
            return False
        logging.warning("Record is not grib2")
        return False

    def is_accumulated(self):
        """Check if accumulated."""
        if self.type_of_statistical_processing == 1:
            logging.debug("Parameter is accumulated")
            return True
        return False

    def print_keys(self):
        """Print keys."""
        logging.info("Version: %s", self.version)
        logging.info("discipline: %s", self.discipline)
        logging.info("parameterCategory: %s", self.parameter_category)
        logging.info("parameterNumber: %s", self.parameter_number)
        logging.info("levelType: %s", self.level_type)
        logging.info("level: %s", self.level)
        logging.info(
            "typeOfStatisticalProcessing: %s", self.type_of_statistical_processing
        )

    def generate_grib_id(self):
        """Generate grib2 ID."""
        dis = str(self.discipline)
        pca = str(self.parameter_category)
        pnr = str(self.parameter_number)
        typ = str(self.level_type)
        lev = str(self.level)
        tsp = str(self.type_of_statistical_processing)
        return f":grib2:{dis}:{pca}:{pnr}:{typ}:{lev}:{tsp}"
